#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ..utils import validate_axis
from ..datasource import tensor as astensor


def compress(condition, a, axis=None):
    """
    Return selected slices of a tensor along given axis.

    When working along a given axis, a slice along that axis is returned in
    `output` for each index where `condition` evaluates to True. When
    working on a 1-D array, `compress` is equivalent to `extract`.

    Parameters
    ----------
    condition : 1-D tensor of bools
        Tensor that selects which entries to return. If len(condition)
        is less than the size of `a` along the given axis, then output is
        truncated to the length of the condition tensor.
    a : array_like
        Tensor from which to extract a part.
    axis : int, optional
        Axis along which to take slices. If None (default), work on the
        flattened tensor.
    out : Tensor, optional
        Output tensor.  Its type is preserved and it must be of the right
        shape to hold the output.

    Returns
    -------
    compressed_array : Tensor
        A copy of `a` without the slices along axis for which `condition`
        is false.

    See Also
    --------
    take, choose, diag, diagonal, select
    Tensor.compress : Equivalent method in ndarray
    mt.extract: Equivalent method when working on 1-D arrays

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array([[1, 2], [3, 4], [5, 6]])
    >>> a.execute()
    array([[1, 2],
           [3, 4],
           [5, 6]])
    >>> mt.compress([0, 1], a, axis=0).execute()
    array([[3, 4]])
    >>> mt.compress([False, True, True], a, axis=0).execute()
    array([[3, 4],
           [5, 6]])
    >>> mt.compress([False, True], a, axis=1).execute()
    array([[2],
           [4],
           [6]])

    Working on the flattened tensor does not return slices along an axis but
    selects elements.

    >>> mt.compress([False, True], a).execute()
    array([2])

    """
    a = astensor(a)
    condition = astensor(condition, dtype=bool)

    if condition.ndim != 1:
        raise ValueError('condition must be an 1-d tensor')

    if axis is None:
        a = a.ravel()
        if len(condition) < a.size:
            a = a[:len(condition)]
        return a[condition]

    try:
        axis = validate_axis(a.ndim, axis)
    except ValueError:
        raise np.AxisError('axis {0} is out of bounds '
                           'for tensor of dimension {1}'.format(axis, a.ndim))

    try:
        if len(condition) < a.shape[axis]:
            a = a[(slice(None),) * axis + (slice(len(condition)),)]
        return a[(slice(None),) * axis + (condition,)]
    except IndexError:
        raise np.AxisError('axis {0} is out of bounds '
                           'for tensor of dimension 1'.format(len(condition)))



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys
import time

from .. import promise
from ..config import options
from ..serialize import dataserializer
from ..utils import mod_hash, log_unhandled, calc_data_size
from ..errors import StoreFull, StoreKeyExists, SpillNotConfigured
from .utils import WorkerActor

logger = logging.getLogger(__name__)


def parse_spill_dirs(dir_str):
    """
    Parse paths from a:b to list while resolving asterisks in path
    """
    import glob
    final_dirs = []
    for pattern in dir_str.split(os.path.pathsep):
        sub_patterns = pattern.split(os.path.sep)
        if not sub_patterns:
            continue
        pos = 0
        while pos < len(sub_patterns) and '*' not in sub_patterns[pos]:
            pos += 1
        if pos == len(sub_patterns):
            final_dirs.append(pattern)
            continue
        left_pattern = os.path.sep.join(sub_patterns[:pos + 1])
        for match in glob.glob(left_pattern):
            final_dirs.append(os.path.sep.join([match] + sub_patterns[pos + 1:]))
    return sorted(final_dirs)


def build_spill_file_name(chunk_key, dirs=None, writing=False):
    """
    Build spill file name from chunk key. Path is selected given hash of the chunk key
    :param chunk_key: chunk key
    """
    dirs = dirs or options.worker.spill_directory
    if not dirs:
        return None
    if not isinstance(dirs, list):
        dirs = parse_spill_dirs(dirs)
    spill_dir = dirs[mod_hash(chunk_key, len(dirs))]
    if writing:
        spill_dir = os.path.join(spill_dir, 'writing')
    if not os.path.exists(spill_dir):
        os.makedirs(spill_dir)
    return os.path.join(spill_dir, chunk_key)


def read_spill_file(chunk_key):
    """
    Read spill file of chunk key via gevent thread pool input_pool
    :param chunk_key: chunk key
    :return: mars object
    """
    file_name = build_spill_file_name(chunk_key)
    if not file_name:
        raise SpillNotConfigured('Spill not configured')
    with open(file_name, 'rb') as file_obj:
        return dataserializer.load(file_obj)


def write_spill_file(chunk_key, data):
    """
    Write mars object into spill
    :param chunk_key: chunk key
    :param data: data of the chunk
    """
    src_file_name = build_spill_file_name(chunk_key, writing=True)
    dest_file_name = build_spill_file_name(chunk_key, writing=False)
    if not src_file_name:
        raise SpillNotConfigured('Spill not configured')
    if not os.path.exists(dest_file_name):
        with open(src_file_name, 'wb') as file_obj:
            dataserializer.dump(data, file_obj, dataserializer.COMPRESS_FLAG_LZ4)
        os.rename(src_file_name, dest_file_name)


def spill_exists(chunk_key):
    file_name = build_spill_file_name(chunk_key)
    if not file_name:
        return False
    return os.path.exists(file_name)


class SpillActor(WorkerActor):
    """
    Actor handling spill read and write in single disk partition
    """
    def __init__(self):
        super(SpillActor, self).__init__()
        if not isinstance(options.worker.spill_directory, list):
            options.worker.spill_directory = options.worker.spill_directory.split(os.path.pathsep)
        self._spill_dirs = options.worker.spill_directory
        self._chunk_holder_ref = None
        self._mem_quota_ref = None
        self._status_ref = None
        self._dispatch_ref = None

        self._input_pool = None
        self._compress_pool = None
        self._output_pool = None

    def post_create(self):
        from .chunkholder import ChunkHolderActor
        from .quota import MemQuotaActor
        from .status import StatusActor
        from .dispatcher import DispatchActor

        super(SpillActor, self).post_create()
        self._chunk_holder_ref = self.promise_ref(ChunkHolderActor.default_name())
        self._mem_quota_ref = self.promise_ref(MemQuotaActor.default_name())

        self._status_ref = self.ctx.actor_ref(StatusActor.default_name())
        if not self.ctx.has_actor(self._status_ref):
            self._status_ref = None

        self._dispatch_ref = self.promise_ref(DispatchActor.default_name())
        self._dispatch_ref.register_free_slot(self.uid, 'spill')

        self._input_pool = self.ctx.threadpool(1)
        self._compress_pool = self.ctx.threadpool(1)
        self._output_pool = self.ctx.threadpool(1)

    @promise.reject_on_exception
    @log_unhandled
    def spill(self, session_id, chunk_key, callback):
        """
        Spill chunk to disk
        :param session_id: session id
        :param chunk_key: chunk key
        :param callback: promise callback
        """
        file_name = build_spill_file_name(chunk_key)
        if not file_name:
            raise SpillNotConfigured('Spill not configured')
        if not os.path.exists(file_name):
            data = None
            try:
                data = self._chunk_store.get(session_id, chunk_key)

                logger.debug('Start spilling chunk %s in %s', chunk_key, self.uid)
                start_time = time.time()
                self._output_pool.submit(write_spill_file, chunk_key, data).result()

                if self._status_ref:
                    self._status_ref.update_mean_stats(
                        'disk_write_speed', calc_data_size(data) * 1.0 / (time.time() - start_time),
                        _tell=True, _wait=False)
            except KeyError:
                if self._chunk_holder_ref.is_stored(chunk_key):
                    raise
            finally:
                del data
        self.tell_promise(callback, file_name)

    @promise.reject_on_exception
    @log_unhandled
    def load(self, session_id, chunk_key, callback=None):
        """
        Load spilled chunk from disk into shared storage
        :param session_id: session id
        :param chunk_key: chunk key
        :param callback: promise callback
        """
        if self._chunk_store.contains(session_id, chunk_key):
            # if already loaded, just register
            ref = None
            try:
                ref = self._chunk_store.get(session_id, chunk_key)
                self._chunk_holder_ref.register_chunk(session_id, chunk_key)
                logger.debug('Chunk %s already loaded in plasma', chunk_key)
                self.tell_promise(callback)
                return
            except KeyError:
                pass
            finally:
                del ref

        logger.debug('Start loading chunk %s from spill', chunk_key)
        file_name = build_spill_file_name(chunk_key)
        if not file_name:
            raise SpillNotConfigured('Spill not configured')

        with open(file_name, 'rb') as inpf:
            data_size = dataserializer.peek_serialized_size(inpf)

        @log_unhandled
        def _handle_rejection(*exc):
            self.tell_promise(callback, *exc, **dict(_accept=False))

        @log_unhandled
        def _try_put_chunk(spill_times=1):
            buf = None
            sealed = False
            try:
                start_time = time.time()
                logger.debug('Creating data writer for chunk %s in plasma', chunk_key)
                buf = self._chunk_store.create(session_id, chunk_key, data_size)
                logger.debug('Successfully created data writer for chunk %s in plasma', chunk_key)

                with open(file_name, 'rb') as inpf,\
                        dataserializer.DecompressBufferWriter(buf) as writer:
                    block_size = options.worker.transfer_block_size
                    compress_future = None
                    while True:
                        read_future = self._input_pool.submit(inpf.read, block_size)
                        if compress_future is not None:
                            compress_future.result()
                        if not read_future.result():
                            break
                        compress_future = self._compress_pool.submit(writer.write, read_future.value)

                self._chunk_store.seal(session_id, chunk_key)
                sealed = True
                if self._status_ref:
                    self._status_ref.update_mean_stats(
                        'disk_read_speed', data_size * 1.0 / (time.time() - start_time),
                        _tell=True, _wait=False)

                self._chunk_holder_ref.register_chunk(session_id, chunk_key)
                logger.debug('Chunk %s loaded from spill to plasma', chunk_key)
                self.tell_promise(callback)
            except (KeyError, StoreKeyExists):
                if not self._chunk_holder_ref.is_stored(chunk_key):
                    raise
                logger.debug('Chunk %s already loaded into plasma', chunk_key)
                self.tell_promise(callback)
            except StoreFull:
                # failed to put into shared cache: spill and retry
                logger.debug('Failed to create data writer for chunk %s in plasma', chunk_key)
                self._chunk_holder_ref.spill_size(data_size, spill_times, _promise=True) \
                    .then(lambda *_: _try_put_chunk(min(spill_times + 1, 1024))) \
                    .catch(_handle_rejection)
            finally:
                if buf is not None and not sealed:
                    self._chunk_store.seal(session_id, chunk_key)
                del buf
        _try_put_chunk()

    @promise.reject_on_exception
    @log_unhandled
    def delete(self, chunk_key):
        """
        Delete chunk data in spill
        :param chunk_key: chunk key
        """
        file_name = build_spill_file_name(chunk_key)
        if not file_name:
            raise SpillNotConfigured('Spill not configured')
        if sys.platform == 'win32':
            CREATE_NO_WINDOW = 0x08000000
            self.ctx.popen(['del', file_name], creationflags=CREATE_NO_WINDOW)
        else:
            self.ctx.popen(['rm', '-f', file_name])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .... import operands
from ...core import Tensor
from ..core import TensorOperandMixin
from ..datasource import tensor as astensor
from .tensordot import tensordot


class TensorDot(operands.Dot, TensorOperandMixin):
    def __init__(self, dtype=None, sparse=False, **kw):
        super(TensorDot, self).__init__(_dtype=dtype, _sparse=sparse, **kw)

    def _set_inputs(self, inputs):
        super(TensorDot, self)._set_inputs(inputs)
        self._a, self._b = self._inputs


def dot(a, b, out=None, sparse=None):
    """
    Dot product of two arrays. Specifically,

    - If both `a` and `b` are 1-D arrays, it is inner product of vectors
      (without complex conjugation).

    - If both `a` and `b` are 2-D arrays, it is matrix multiplication,
      but using :func:`matmul` or ``a @ b`` is preferred.

    - If either `a` or `b` is 0-D (scalar), it is equivalent to :func:`multiply`
      and using ``numpy.multiply(a, b)`` or ``a * b`` is preferred.

    - If `a` is an N-D array and `b` is a 1-D array, it is a sum product over
      the last axis of `a` and `b`.

    - If `a` is an N-D array and `b` is an M-D array (where ``M>=2``), it is a
      sum product over the last axis of `a` and the second-to-last axis of `b`::

        dot(a, b)[i,j,k,m] = sum(a[i,j,:] * b[k,:,m])

    Parameters
    ----------
    a : array_like
        First argument.
    b : array_like
        Second argument.
    out : Tensor, optional
        Output argument. This must have the exact kind that would be returned
        if it was not used. In particular, it must have the right type, must be
        C-contiguous, and its dtype must be the dtype that would be returned
        for `dot(a,b)`. This is a performance feature. Therefore, if these
        conditions are not met, an exception is raised, instead of attempting
        to be flexible.

    Returns
    -------
    output : Tensor
        Returns the dot product of `a` and `b`.  If `a` and `b` are both
        scalars or both 1-D arrays then a scalar is returned; otherwise
        a tensor is returned.
        If `out` is given, then it is returned.

    Raises
    ------
    ValueError
        If the last dimension of `a` is not the same size as
        the second-to-last dimension of `b`.

    See Also
    --------
    vdot : Complex-conjugating dot product.
    tensordot : Sum products over arbitrary axes.
    einsum : Einstein summation convention.
    matmul : '@' operator as method with out parameter.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.dot(3, 4).execute()
    12

    Neither argument is complex-conjugated:

    >>> mt.dot([2j, 3j], [2j, 3j]).execute()
    (-13+0j)

    For 2-D arrays it is the matrix product:

    >>> a = [[1, 0], [0, 1]]
    >>> b = [[4, 1], [2, 2]]
    >>> mt.dot(a, b).execute()
    array([[4, 1],
           [2, 2]])

    >>> a = mt.arange(3*4*5*6).reshape((3,4,5,6))
    >>> b = mt.arange(3*4*5*6)[::-1].reshape((5,4,6,3))
    >>> mt.dot(a, b)[2,3,2,1,2,2].execute()
    499128
    >>> mt.sum(a[2,3,2,:] * b[1,2,:,2]).execute()
    499128
    """
    a, b = astensor(a), astensor(b)
    if a.isscalar() and b.isscalar():
        ret = a * b
    else:
        ret = tensordot(a, b, axes=((a.ndim - 1,), (b.ndim - 2,)), sparse=sparse)

    if out is None:
        return ret

    # set to out
    if not isinstance(out, Tensor):
        raise ValueError('`out` must be a Tensor, got {0} instead'.format(type(out)))
    out.data = ret.data
    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..arithmetic.sqrt import sqrt
from .nanvar import nanvar


def nanstd(a, axis=None, dtype=None, out=None, ddof=0, keepdims=None, combine_size=None):
    """
    Compute the standard deviation along the specified axis, while
    ignoring NaNs.

    Returns the standard deviation, a measure of the spread of a
    distribution, of the non-NaN tensor elements. The standard deviation is
    computed for the flattened tensor by default, otherwise over the
    specified axis.

    For all-NaN slices or slices with zero degrees of freedom, NaN is
    returned and a `RuntimeWarning` is raised.

    Parameters
    ----------
    a : array_like
        Calculate the standard deviation of the non-NaN values.
    axis : int, optional
        Axis along which the standard deviation is computed. The default is
        to compute the standard deviation of the flattened tensor.
    dtype : dtype, optional
        Type to use in computing the standard deviation. For tensors of
        integer type the default is float64, for tensors of float types it
        is the same as the tensor type.
    out : Tensor, optional
        Alternative output tensor in which to place the result. It must have
        the same shape as the expected output but the type (of the
        calculated values) will be cast if necessary.
    ddof : int, optional
        Means Delta Degrees of Freedom.  The divisor used in calculations
        is ``N - ddof``, where ``N`` represents the number of non-NaN
        elements.  By default `ddof` is zero.

    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one. With this option,
        the result will broadcast correctly against the original `a`.

        If this value is anything but the default it is passed through
        as-is to the relevant functions of the sub-classes.  If these
        functions do not have a `keepdims` kwarg, a RuntimeError will
        be raised.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    standard_deviation : ndarray, see dtype parameter above.
        If `out` is None, return a new array containing the standard
        deviation, otherwise return a reference to the output tensor. If
        ddof is >= the number of non-NaN elements in a slice or the slice
        contains only NaNs, then the result for that slice is NaN.

    See Also
    --------
    var, mean, std
    nanvar, nanmean

    Notes
    -----
    The standard deviation is the square root of the average of the squared
    deviations from the mean: ``std = sqrt(mean(abs(x - x.mean())**2))``.

    The average squared deviation is normally calculated as
    ``x.sum() / N``, where ``N = len(x)``.  If, however, `ddof` is
    specified, the divisor ``N - ddof`` is used instead. In standard
    statistical practice, ``ddof=1`` provides an unbiased estimator of the
    variance of the infinite population. ``ddof=0`` provides a maximum
    likelihood estimate of the variance for normally distributed variables.
    The standard deviation computed in this function is the square root of
    the estimated variance, so even with ``ddof=1``, it will not be an
    unbiased estimate of the standard deviation per se.

    Note that, for complex numbers, `std` takes the absolute value before
    squaring, so that the result is always real and nonnegative.

    For floating-point input, the *std* is computed using the same
    precision the input has. Depending on the input data, this can cause
    the results to be inaccurate, especially for float32 (see example
    below).  Specifying a higher-accuracy accumulator using the `dtype`
    keyword can alleviate this issue.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array([[1, mt.nan], [3, 4]])
    >>> mt.nanstd(a).execute()
    1.247219128924647
    >>> mt.nanstd(a, axis=0).execute()
    array([ 1.,  0.])
    >>> mt.nanstd(a, axis=1).execute()
    array([ 0.,  0.5])

    """
    ret = sqrt(nanvar(a, axis=axis, dtype=dtype, out=out, ddof=ddof, keepdims=keepdims,
                      combine_size=combine_size))
    if dtype is not None and ret.dtype != dtype:
        ret = ret.astype(dtype)
    return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import opcodes as OperandDef
from ....serialize import Int64Field, BoolField
from ....config import options
from ..utils import decide_chunks
from .core import TensorNoInput


class TensorLinspace(TensorNoInput):
    _op_type_ = OperandDef.TENSOR_LINSPACE

    _start = Int64Field('start')
    _stop = Int64Field('stop')
    _num = Int64Field('num')
    _endpoint = BoolField('endpoint')

    def __init__(self, start=None, stop=None, num=None, endpoint=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(np.linspace(0, 1, 1).dtype if dtype is None else dtype)
        super(TensorLinspace, self).__init__(_start=start, _stop=stop, _num=num,
                                             _endpoint=endpoint, _dtype=dtype, _gpu=gpu, **kw)

    def to_chunk_op(self, *args):
        start, stop, num, endpoint = args
        op = self.copy().reset_key()
        op._start = start
        op._stop = stop
        op._num = num
        op._endpoint = endpoint
        return op

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]

        chunk_length = tensor.params.raw_chunks or options.tensor.chunks
        chunk_length = decide_chunks(tensor.shape, chunk_length, tensor.dtype.itemsize)

        start, stop, num, endpoint = \
            tensor.op.start, tensor.op.stop, tensor.op.num, tensor.op.endpoint
        if num > 1:
            step = float(stop - start) / (num if not endpoint else num - 1)
        else:
            step = 0.0

        chunks = []
        chunk_start = start
        nsplit = []
        for i, cs in enumerate(chunk_length[0]):
            chunk_stop = chunk_start + (cs - 1) * step
            chunk_op = op.to_chunk_op(chunk_start, chunk_stop, cs, True)
            chunk_shape = (cs,)
            chunk_idx = (i,)
            chunk = chunk_op.new_chunk(None, chunk_shape, index=chunk_idx)
            chunks.append(chunk)
            nsplit.append(cs)
            chunk_start = chunk_start + cs * step

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, op.outputs[0].shape, chunks=chunks,
                                  nsplits=(tuple(nsplit),))

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    @property
    def num(self):
        return self._num

    @property
    def endpoint(self):
        return self._endpoint


def linspace(start, stop, num=50, endpoint=True, retstep=False,
             dtype=None, gpu=False, chunks=None):
    """
    Return evenly spaced numbers over a specified interval.

    Returns `num` evenly spaced samples, calculated over the
    interval [`start`, `stop`].

    The endpoint of the interval can optionally be excluded.

    Parameters
    ----------
    start : scalar
        The starting value of the sequence.
    stop : scalar
        The end value of the sequence, unless `endpoint` is set to False.
        In that case, the sequence consists of all but the last of ``num + 1``
        evenly spaced samples, so that `stop` is excluded.  Note that the step
        size changes when `endpoint` is False.
    num : int, optional
        Number of samples to generate. Default is 50. Must be non-negative.
    endpoint : bool, optional
        If True, `stop` is the last sample. Otherwise, it is not included.
        Default is True.
    retstep : bool, optional
        If True, return (`samples`, `step`), where `step` is the spacing
        between samples.
    dtype : dtype, optional
        The type of the output tensor.  If `dtype` is not given, infer the data
        type from the other input arguments.
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    chunks : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension

    Returns
    -------
    samples : Tensor
        There are `num` equally spaced samples in the closed interval
        ``[start, stop]`` or the half-open interval ``[start, stop)``
        (depending on whether `endpoint` is True or False).
    step : float, optional
        Only returned if `retstep` is True

        Size of spacing between samples.


    See Also
    --------
    arange : Similar to `linspace`, but uses a step size (instead of the
             number of samples).
    logspace : Samples uniformly distributed in log space.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.session import new_session

    >>> sess = new_session().as_default()

    >>> mt.linspace(2.0, 3.0, num=5).execute()
    array([ 2.  ,  2.25,  2.5 ,  2.75,  3.  ])
    >>> mt.linspace(2.0, 3.0, num=5, endpoint=False).execute()
    array([ 2. ,  2.2,  2.4,  2.6,  2.8])
    >>> sess.run(mt.linspace(2.0, 3.0, num=5, retstep=True).execute())
    (array([ 2.  ,  2.25,  2.5 ,  2.75,  3.  ]), 0.25)

    Graphical illustration:

    >>> import matplotlib.pyplot as plt
    >>> N = 8
    >>> y = mt.zeros(N)
    >>> x1 = mt.linspace(0, 10, N, endpoint=True)
    >>> x2 = mt.linspace(0, 10, N, endpoint=False)
    >>> plt.plot(x1.execute(), y.execute(), 'o')
    [<matplotlib.lines.Line2D object at 0x...>]
    >>> plt.plot(x2.execute(), y.execute() + 0.5, 'o')
    [<matplotlib.lines.Line2D object at 0x...>]
    >>> plt.ylim([-0.5, 1])
    (-0.5, 1)
    >>> plt.show()

    """
    num = int(num)

    op = TensorLinspace(start, stop, num, endpoint, dtype=dtype, gpu=gpu)
    shape = (num,)
    ret = op(shape, chunks=chunks)

    if not retstep:
        return ret

    if num > 1:
        step = float(stop - start) / (num if not endpoint else num - 1)
    else:
        step = np.nan

    return ret, step



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .split import _split


def array_split(a, indices_or_sections, axis=0):
    """
    Split a tensor into multiple sub-tensors.

    Please refer to the ``split`` documentation.  The only difference
    between these functions is that ``array_split`` allows
    `indices_or_sections` to be an integer that does *not* equally
    divide the axis. For a tensor of length l that should be split
    into n sections, it returns l % n sub-arrays of size l//n + 1
    and the rest of size l//n.

    See Also
    --------
    split : Split tensor into multiple sub-tensors of equal size.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> from mars.session import new_session

    >>> sess = new_session().as_default()

    >>> x = mt.arange(8.0)
    >>> sess.run(mt.array_split(x, 3))
        [array([ 0.,  1.,  2.]), array([ 3.,  4.,  5.]), array([ 6.,  7.])]

    >>> x = mt.arange(7.0)
    >>> sess.run(mt.array_split(x, 3))
        [array([ 0.,  1.,  2.]), array([ 3.,  4.]), array([ 5.,  6.])]

    """
    return _split(a, indices_or_sections, axis=axis)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys
import time
import zlib
from collections import defaultdict, OrderedDict

from .. import promise
from ..compat import six, Enum
from ..config import options
from ..serialize import dataserializer
from ..errors import *
from ..utils import log_unhandled
from .spill import build_spill_file_name
from .utils import WorkerActor

logger = logging.getLogger(__name__)


class ReceiveStatus(Enum):
    NOT_STARTED = 0
    RECEIVING = 1
    RECEIVED = 2
    ERROR = 2


class SenderActor(WorkerActor):
    """
    Actor handling sending data to ReceiverActors in other workers
    """
    def __init__(self):
        super(SenderActor, self).__init__()
        self._dispatch_ref = None
        self._mem_quota_ref = None

        self._serialize_pool = None

    def post_create(self):
        from .dispatcher import DispatchActor
        from .quota import MemQuotaActor

        super(SenderActor, self).post_create()
        self._dispatch_ref = self.promise_ref(DispatchActor.default_name())
        self._dispatch_ref.register_free_slot(self.uid, 'sender')
        self._mem_quota_ref = self.promise_ref(MemQuotaActor.default_name())

        self._serialize_pool = self.ctx.threadpool(1)

    def _read_data_size(self, session_id, chunk_key):
        """
        Obtain decompressed data size for a certain chunk
        :param session_id: session id
        :param chunk_key: chunk key
        :return: size of data
        """
        nbytes = None
        if self._chunk_store.contains(session_id, chunk_key):
            try:
                nbytes = self._chunk_store.get_actual_size(session_id, chunk_key)
            except KeyError:
                pass
        if nbytes is None:
            file_name = build_spill_file_name(chunk_key)
            if not file_name:
                raise SpillNotConfigured('Spill not configured')
            if not os.path.exists(file_name):
                raise DependencyMissing('Dependency %s not met on sending.' % chunk_key)
            with open(file_name, 'rb') as inf:
                nbytes = self._serialize_pool.submit(dataserializer.peek_serialized_size, inf).result()
        return nbytes

    @promise.reject_on_exception
    @log_unhandled
    def send_data(self, session_id, chunk_key, target_endpoints, ensure_cached=True,
                  timeout=0, callback=None):
        """
        Send data to other workers
        :param session_id: session id
        :param chunk_key: chunk to be sent
        :param target_endpoints: endpoints to receive this chunk
        :param ensure_cached: if True, make sure the data is in the shared storage of the target worker
        :param timeout: timeout of data sending
        :param callback: promise callback
        """
        from .dispatcher import DispatchActor

        remote_receiver_refs = []
        already_started = set()
        finish_promises = []
        if isinstance(target_endpoints, six.string_types):
            target_endpoints = [target_endpoints]

        try:
            logger.debug('Begin sending data %s into endpoints %s', chunk_key, target_endpoints)
            # collect receiver actors and quota actors in remote workers
            for ep in target_endpoints:
                dispatch_ref = self.promise_ref(DispatchActor.default_name(), address=ep)
                uid = dispatch_ref.get_hash_slot('receiver', chunk_key)

                receiver_ref = self.promise_ref(uid, address=ep)
                remote_status = receiver_ref.check_status(session_id, chunk_key)
                if remote_status == ReceiveStatus.RECEIVED:
                    # data already been sent, no need to transfer any more
                    continue
                elif remote_status == ReceiveStatus.RECEIVING:
                    # data under transfer, we only need to listen to the transfer progress
                    finish_promises.append(
                        receiver_ref.register_finish_callback(session_id, chunk_key, _promise=True))
                    continue
                remote_receiver_refs.append(receiver_ref)

            if not remote_receiver_refs and not finish_promises:
                # no transfer needed, we exit and release slot resource
                logger.debug('No data needed to transfer for chunk %s, invoke callback directly', chunk_key)
                self._dispatch_ref.register_free_slot(self.uid, 'sender', _tell=True)
                self.tell_promise(callback, None)
                return
            data_size = self._read_data_size(session_id, chunk_key)
        except:
            self._dispatch_ref.register_free_slot(self.uid, 'sender', _tell=True)
            raise

        @log_unhandled
        def _handle_creation(address, status):
            # filter out endpoints already transferred or already started transfer
            if status in (ReceiveStatus.RECEIVING, ReceiveStatus.RECEIVED):
                already_started.add(address)

        @log_unhandled
        def _compress_and_send():
            # start compress and send data into targets
            logger.debug('Data writer for chunk %s allocated at targets, start transmission', chunk_key)
            min_chunk_size = options.worker.transfer_block_size
            reader = None

            # filter out endpoints we need to send to
            active_receiver_refs = [ref for ref in remote_receiver_refs
                                    if ref.address not in already_started]
            try:
                if not active_receiver_refs:
                    self._dispatch_ref.register_free_slot(self.uid, 'sender', _tell=True)
                    return

                # try load data from plasma store
                if self._chunk_store.contains(session_id, chunk_key):
                    buf = None
                    try:
                        buf = self._chunk_store.get_buffer(session_id, chunk_key)
                        # create a stream compressor from shared buffer
                        reader = dataserializer.CompressBufferReader(buf, dataserializer.COMPRESS_FLAG_LZ4)
                    except KeyError:
                        pass
                    finally:
                        del buf
                if reader is None:
                    # no reader created from plasma store, we load directly from spill
                    file_name = build_spill_file_name(chunk_key)
                    if not file_name:
                        raise SpillNotConfigured('Spill not configured')
                    reader = open(file_name, 'rb')

                futures = []
                checksum = 0
                while True:
                    # read a data part from reader we defined above
                    next_chunk = self._serialize_pool.submit(reader.read, min_chunk_size).result()
                    # make sure all previous transfers finished
                    [f.result() for f in futures]
                    if not next_chunk:
                        # no further data to read, we close and finish the transfer
                        reader.close()
                        for ref in active_receiver_refs:
                            ref.finish_receive(session_id, chunk_key, checksum, _tell=True)
                        break
                    checksum = zlib.crc32(next_chunk, checksum)
                    futures = []
                    for ref in active_receiver_refs:
                        # we perform async transfer and wait after next part is loaded and compressed
                        futures.append(ref.receive_data_part(
                            session_id, chunk_key, next_chunk, checksum, _wait=False))
            except:
                for ref in active_receiver_refs:
                    ref.cancel_receive(session_id, chunk_key)
                raise
            finally:
                if reader:
                    reader.close()
                del reader

        @log_unhandled
        def _finalize(*_):
            self._dispatch_ref.register_free_slot(self.uid, 'sender', _tell=True)
            self.tell_promise(callback, data_size)

        @log_unhandled
        def _handle_rejection(*exc):
            for ref in remote_receiver_refs:
                ref.cancel_receive(session_id, chunk_key, _tell=True)
            self._dispatch_ref.register_free_slot(self.uid, 'sender', _tell=True)
            self.tell_promise(callback, *exc, **dict(_accept=False))

        promises = []
        for ref in remote_receiver_refs:
            # register transfer actions
            promises.append(
                ref.create_data_writer(
                    session_id, chunk_key, data_size, self.ref(), ensure_cached=ensure_cached,
                    _timeout=timeout, _promise=True
                ).then(_handle_creation)
            )
            # register finish listener
            finish_promises.append(ref.register_finish_callback(session_id, chunk_key, _promise=True))

        if promises:
            promise.all_(promises).then(_compress_and_send).catch(_handle_rejection)
        else:
            # nothing to send, the slot can be released
            self._dispatch_ref.register_free_slot(self.uid, 'sender', _tell=True)
        # wait for all finish listeners returned
        promise.all_(finish_promises).then(_finalize).catch(_handle_rejection)


class ReceiverActor(WorkerActor):
    """
    Actor handling receiving data from a SenderActor
    """
    def __init__(self):
        super(ReceiverActor, self).__init__()
        self._chunk_holder_ref = None
        self._mem_quota_ref = None
        self._dispatch_ref = None
        self._status_ref = None

        self._finish_callbacks = defaultdict(list)
        self._data_writers = dict()
        self._data_meta_cache = OrderedDict()

        self._serialize_pool = None

    def post_create(self):
        from .chunkholder import ChunkHolderActor
        from .quota import MemQuotaActor
        from .status import StatusActor
        from .dispatcher import DispatchActor

        super(ReceiverActor, self).post_create()
        self._chunk_holder_ref = self.promise_ref(ChunkHolderActor.default_name())
        self._mem_quota_ref = self.promise_ref(MemQuotaActor.default_name())

        self._status_ref = self.ctx.actor_ref(StatusActor.default_name())
        if not self.ctx.has_actor(self._status_ref):
            self._status_ref = None

        self._dispatch_ref = self.promise_ref(DispatchActor.default_name())
        self._dispatch_ref.register_free_slot(self.uid, 'receiver')

        self._serialize_pool = self.ctx.threadpool(1)

    @log_unhandled
    def check_status(self, session_id, chunk_key):
        """
        Check if data exists or is being transferred in the target worker
        :param session_id: session id
        :param chunk_key: chunk key
        """
        from .spill import build_spill_file_name
        session_chunk_key = (session_id, chunk_key)

        if self._chunk_store.contains(session_id, chunk_key):
            # data in plasma
            return ReceiveStatus.RECEIVED
        if session_chunk_key in self._data_writers:
            # data still being transferred
            return ReceiveStatus.RECEIVING
        fn = build_spill_file_name(chunk_key)
        if fn and os.path.exists(fn) and session_chunk_key not in self._data_writers:
            # data in plasma store
            return ReceiveStatus.RECEIVED
        return ReceiveStatus.NOT_STARTED

    @promise.reject_on_exception
    @log_unhandled
    def register_finish_callback(self, session_id, chunk_key, callback):
        """
        Register a promise callback to handle transfer termination.
        If the chunk has already been transferred or error occurred,
        the existing results are sent.
        :param session_id: session id
        :param chunk_key: chunk key
        :param callback: promise callback
        """
        session_chunk_key = (session_id, chunk_key)
        if session_chunk_key not in self._data_meta_cache:
            self._finish_callbacks[session_chunk_key].append(callback)
            return
        data_meta = self._data_meta_cache[session_chunk_key]
        if data_meta['status'] in (ReceiveStatus.RECEIVED, ReceiveStatus.ERROR):
            # invoke callback directly when transfer finishes
            self.tell_promise(callback, *data_meta['callback_args'], **data_meta['callback_kwargs'])
        else:
            self._finish_callbacks[session_chunk_key].append(callback)

    @promise.reject_on_exception
    @log_unhandled
    def create_data_writer(self, session_id, chunk_key, data_size, sender_ref,
                           ensure_cached=True, callback=None):
        """
        Create a data writer for subsequent data transfer. The writer can either work on
        shared storage or spill.
        :param session_id: session id
        :param chunk_key: chunk key
        :param data_size: uncompressed data size
        :param sender_ref: ActorRef of SenderActor
        :param ensure_cached: if True, the data should be stored in shared memory, otherwise spill is acceptable
        :param callback: promise callback
        """
        logger.debug('Begin creating transmission data writer for chunk %s from %s',
                     chunk_key, sender_ref.address)
        session_chunk_key = (session_id, chunk_key)

        data_status = self.check_status(session_id, chunk_key)

        if data_status == ReceiveStatus.RECEIVING:
            # data transfer already started
            logger.debug('Chunk %s already started transmission', chunk_key)
            if callback:
                self.tell_promise(callback, self.address, ReceiveStatus.RECEIVING)
            return

        # build meta data for data transfer
        if session_chunk_key in self._data_meta_cache:
            if data_status == ReceiveStatus.RECEIVED:
                # already received: callback directly
                logger.debug('Chunk %s already started received', chunk_key)
                if callback:
                    self.tell_promise(callback, self.address, ReceiveStatus.RECEIVED)
                self._invoke_finish_callbacks(session_id, chunk_key)
                return
            else:
                del self._data_meta_cache[session_chunk_key]

        self._data_meta_cache[session_chunk_key] = dict(
            transfer_start_time=time.time(),
            chunk_size=data_size,
            write_shared=True,
            checksum=0,
            status=ReceiveStatus.RECEIVING,
            callback_args=(),
            callback_kwargs={},
        )

        @log_unhandled
        def _handle_reject(*exc):
            logger.debug('Rejecting %s from putting into plasma.', chunk_key)
            self._stop_transfer_with_exc(session_id, chunk_key, exc)
            if callback:
                self.tell_promise(callback, *exc, **dict(_accept=False))

        @log_unhandled
        def _create_writer(spill_times=1):
            # actual create data writer
            buf = None
            try:
                # attempt to create data chunk on shared store
                buf = self._chunk_store.create(session_id, chunk_key, data_size)
                self._data_meta_cache[session_chunk_key]['transfer_start_time'] = time.time()
                logger.debug('Successfully created data writer with %s bytes in plasma for chunk %s',
                             data_size, chunk_key)
                # create a writer for the chunk
                self._data_writers[session_chunk_key] = dataserializer.DecompressBufferWriter(buf)
                if callback:
                    self.tell_promise(callback, self.address, None)
            except (KeyError, StoreKeyExists):
                if self.check_status(session_id, chunk_key) != ReceiveStatus.RECEIVED:
                    raise
                # data already registered
                if callback:
                    logger.debug('Chunk %s already registered', chunk_key)
                    self.tell_promise(callback, self.address, ReceiveStatus.RECEIVED)
                    self._invoke_finish_callbacks(session_id, chunk_key)
            except StoreFull:
                # no space left in the shared store
                if ensure_cached:
                    # spill and try again
                    self._chunk_holder_ref.spill_size(data_size, spill_times, _promise=True) \
                        .then(lambda *_: _create_writer(min(spill_times + 1, 1024))) \
                        .catch(_handle_reject)
                else:
                    # create a writer for spill
                    logger.debug('Writing data %s directly into spill.', chunk_key)
                    self._data_meta_cache[session_chunk_key]['write_shared'] = False
                    self._chunk_holder_ref.spill_size(data_size, _tell=True)
                    spill_file_name = build_spill_file_name(chunk_key, writing=True)
                    try:
                        spill_file = open(spill_file_name, 'wb')
                        self._data_meta_cache[session_chunk_key]['transfer_start_time'] = time.time()
                        self._data_writers[session_chunk_key] = spill_file
                    except (KeyError, IOError):
                        if self.check_status(session_id, chunk_key) == ReceiveStatus.RECEIVED:
                            logger.debug('Chunk %s already stored', chunk_key)
                            if callback:
                                self.tell_promise(callback, self.address, ReceiveStatus.RECEIVED)
                            self._invoke_finish_callbacks(session_id, chunk_key)
                            return
                        raise ObjectNotInPlasma([chunk_key])
                    if callback:
                        self.tell_promise(callback, self.address, None)
            finally:
                del buf

        promise.Promise(done=True) \
            .then(lambda *_: _create_writer()) \
            .catch(_handle_reject)

    @log_unhandled
    def receive_data_part(self, session_id, chunk_key, data_part, checksum):
        """
        Receive data part from sender
        :param session_id: session id
        :param chunk_key: chunk key
        :param data_part: data to be written
        :param checksum: checksum up to now
        """
        session_chunk_key = (session_id, chunk_key)
        try:
            data_meta = self._data_meta_cache[session_chunk_key]

            # check if checksum matches
            local_checksum = zlib.crc32(data_part, data_meta['checksum'])
            if local_checksum != checksum:
                raise ChecksumMismatch
            data_meta['checksum'] = local_checksum

            # if error occurred, interrupts
            if data_meta['status'] == ReceiveStatus.ERROR:
                six.reraise(*self._data_meta_cache[session_chunk_key]['callback_args'])
                return
            self._serialize_pool.submit(self._data_writers[session_chunk_key].write, data_part).result()
        except:
            self._stop_transfer_with_exc(session_id, chunk_key, sys.exc_info())

    @log_unhandled
    def finish_receive(self, session_id, chunk_key, checksum):
        """
        Finish data receiving and seal data
        :param session_id: session id
        :param chunk_key: chunk key
        :param checksum: checksum of compressed data
        """
        try:
            logger.debug('Finishing transfer for data %s.', chunk_key)
            session_chunk_key = (session_id, chunk_key)

            data_meta = self._data_meta_cache[session_chunk_key]
            # if checksum mismatch, raises
            if checksum != data_meta['checksum']:
                raise ChecksumMismatch

            # update transfer speed stats
            if self._status_ref:
                time_delta = time.time() - data_meta['transfer_start_time']
                self._status_ref.update_mean_stats(
                    'net_transfer_speed', data_meta['chunk_size'] * 1.0 / time_delta,
                    _tell=True, _wait=False)

            if data_meta['write_shared']:
                # seal data on shared store
                self._chunk_store.seal(session_id, chunk_key)
                self._chunk_holder_ref.register_chunk(session_id, chunk_key)
            else:
                # move spill data to 'ready' place
                src_dir = build_spill_file_name(chunk_key, writing=True)
                dest_dir = build_spill_file_name(chunk_key, writing=False)
                os.rename(src_dir, dest_dir)

            chunk_path = '/sessions/%s/chunks/%s' % (session_id, chunk_key)
            futures = [
                self._kv_store_ref.write(chunk_path + '/workers/%s' % self.address, '', _wait=False),
                self._kv_store_ref.write(chunk_path + '/data_size', data_meta['chunk_size'], _wait=False),
            ]
            [f.result() for f in futures]

            self._data_writers[session_chunk_key].close()
            del self._data_writers[session_chunk_key]

            data_meta['status'] = ReceiveStatus.RECEIVED
            self._invoke_finish_callbacks(session_id, chunk_key)
        except:
            self._stop_transfer_with_exc(session_id, chunk_key, sys.exc_info())

    @log_unhandled
    def cancel_receive(self, session_id, chunk_key):
        """
        Cancel data receive by returning an ExecutionInterrupted
        :param session_id: session id
        :param chunk_key: chunk key
        """
        logger.debug('Transfer for %s cancelled.', chunk_key)
        session_chunk_key = (session_id, chunk_key)
        if session_chunk_key not in self._data_meta_cache:
            return

        if self._data_meta_cache[session_chunk_key]['status'] in (ReceiveStatus.ERROR, ReceiveStatus.RECEIVED):
            # already terminated, we do nothing
            return

        try:
            raise ExecutionInterrupted
        except ExecutionInterrupted:
            self._stop_transfer_with_exc(session_id, chunk_key, sys.exc_info())

    def _stop_transfer_with_exc(self, session_id, chunk_key, exc):
        try:
            six.reraise(*exc)
        except ExecutionInterrupted:
            pass
        except:
            logger.exception('Exception occurred in transferring %s. Cancelling transfer.', chunk_key)

        session_chunk_key = (session_id, chunk_key)

        # stop and close data writer
        if session_chunk_key in self._data_writers:
            self._data_writers[session_chunk_key].close()
            del self._data_writers[session_chunk_key]

        data_meta = self._data_meta_cache[session_chunk_key]
        # clean up unfinished transfers
        if data_meta['write_shared']:
            try:
                self._chunk_store.seal(session_id, chunk_key)
            except KeyError:
                pass
        else:
            src_dir = build_spill_file_name(chunk_key, writing=True)
            if os.path.exists(src_dir):
                os.unlink(src_dir)

        self._data_meta_cache['status'] = ReceiveStatus.ERROR
        self._invoke_finish_callbacks(session_id, chunk_key, *exc, **dict(_accept=False))

    def _invoke_finish_callbacks(self, session_id, chunk_key, *args, **kwargs):
        # invoke registered callbacks for chunk
        session_chunk_key = (session_id, chunk_key)
        data_meta = self._data_meta_cache[session_chunk_key]
        data_meta['callback_args'] = args
        data_meta['callback_kwargs'] = kwargs

        for cb in self._finish_callbacks[session_chunk_key]:
            self.tell_promise(cb, *args, **kwargs)
        if session_chunk_key in self._finish_callbacks:
            del self._finish_callbacks[session_chunk_key]

        # remove outdated metadata
        clean_keys = []
        last_finish_time = time.time() - options.worker.callback_preserve_time
        for k in self._data_meta_cache:
            if self._data_meta_cache[k]['transfer_start_time'] < last_finish_time:
                clean_keys.append(k)
            else:
                break
        for k in clean_keys:
            del self._data_meta_cache[k]


class ResultSenderActor(WorkerActor):
    """
    Actor handling sending result to user client
    """
    def __init__(self):
        super(ResultSenderActor, self).__init__()
        self._serialize_pool = None

    def post_create(self):
        super(ResultSenderActor, self).post_create()
        self._serialize_pool = self.ctx.threadpool(1)

    def fetch_data(self, session_id, chunk_key):
        buf = None
        try:
            if self._chunk_store.contains(session_id, chunk_key):
                buf = self._chunk_store.get_buffer(session_id, chunk_key)
                compressed = self._serialize_pool.submit(
                    dataserializer.dumps, buf, dataserializer.COMPRESS_FLAG_LZ4, raw=True).result()
            else:
                file_name = build_spill_file_name(chunk_key)
                if not file_name:
                    raise SpillNotConfigured('Spill not configured')
                with open(file_name, 'rb') as inf:
                    compressed = self._serialize_pool.submit(inf.read).result()
        finally:
            del buf
        return compressed



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict

from .server import register_ui_handler, get_jinja_env
from ..utils import to_str
from .server import MarsWebAPI

_jinja_env = get_jinja_env()


def worker_list(doc, workers_meta):
    doc.title = 'Mars UI'

    doc.template_variables['worker_metrics'] = workers_meta
    doc.template = _jinja_env.get_template('worker_list.html')


def worker_detail(doc, workers_meta, endpoint):
    doc.title = 'Mars UI'

    progress_data = workers_meta[endpoint]['progress']
    progresses = dict()
    for session_data in progress_data.values():
        for g in session_data.values():
            if g['stage'] not in progresses:
                progresses[g['stage']] = dict(operands=defaultdict(lambda: 0), total=0)
            progresses[g['stage']]['operands'][g['operands']] += 1
            progresses[g['stage']]['total'] += 1
    for k in progresses:
        operands = sorted(progresses[k]['operands'].items(), key=lambda p: p[0])
        progresses[k]['operands'] = ', '.join('%s (%d)' % (k, v) for k, v in operands)

    doc.template_variables['endpoint'] = endpoint
    doc.template_variables['worker_metrics'] = workers_meta[endpoint]
    doc.template_variables['progresses'] = progresses
    doc.template = _jinja_env.get_template('worker_detail.html')


def _route(scheduele_ip, doc):
    web_api = MarsWebAPI(scheduele_ip)
    workers_meta = web_api.get_workers_meta()

    endpoint = doc.session_context.request.arguments.get('endpoint')
    if not endpoint:
        return worker_list(doc, workers_meta)
    else:
        return worker_detail(doc, workers_meta, to_str(endpoint[0]))


register_ui_handler('/worker', _route)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import time

try:
    from pyarrow import plasma
except ImportError:
    plasma = None

from ..config import options
from .. import resource, kvstore
from ..utils import parse_memory_limit, readable_size
from ..compat import six
from ..cluster_info import ClusterInfoActor
from .status import StatusActor
from .quota import QuotaActor, MemQuotaActor
from .chunkholder import ChunkHolderActor
from .dispatcher import DispatchActor
from .execution import ExecutionActor
from .calc import CpuCalcActor
from .transfer import ReceiverActor, SenderActor
from .prochelper import ProcessHelperActor
from .transfer import ResultSenderActor
from .spill import SpillActor


logger = logging.getLogger(__name__)


class WorkerService(object):
    service_logger = logger

    def __init__(self):
        self._plasma_store = None

        self._chunk_holder_ref = None
        self._mem_quota_ref = None
        self._dispatch_ref = None
        self._status_ref = None
        self._execution_ref = None

        self._cluster_info_ref = None
        self._cpu_calc_actors = []
        self._sender_actors = []
        self._receiver_actors = []
        self._spill_actors = []
        self._process_helper_actors = []
        self._result_sender_ref = None

    def start_plasma(self, mem_limit, one_mapped_file=False):
        self._plasma_store = plasma.start_plasma_store(
            int(mem_limit), use_one_memory_mapped_file=one_mapped_file
        )
        options.worker.plasma_socket, _ = self._plasma_store.__enter__()

    @classmethod
    def _calc_soft_memory_limit(cls):
        """
        Calc memory limit for MemQuotaActor given configured percentage. Cache size and memory
        already been used will be excluded from calculated size.
        """
        mem_status = resource.virtual_memory()

        phy_soft_limit = options.worker.physical_memory_limit_soft
        cache_limit = options.worker.cache_memory_limit
        actual_used = mem_status.total - mem_status.available
        quota_soft_limit = phy_soft_limit - cache_limit - actual_used

        if quota_soft_limit < 512 * 1024 ** 2:
            raise MemoryError('Memory not sufficient. soft_limit=%s, cache_limit=%s, used=%s'
                              % tuple(readable_size(k) for k in (phy_soft_limit, cache_limit, actual_used)))

        cls.service_logger.info('Setting soft limit to %s.', readable_size(quota_soft_limit))
        return quota_soft_limit

    def start(self, endpoint, schedulers, pool, ignore_avail_mem=False):
        if schedulers:
            if isinstance(schedulers, six.string_types):
                schedulers = [schedulers]
            service_discover_addr = None
        else:
            schedulers = None
            service_discover_addr = options.kv_store

        # create ClusterInfoActor
        self._cluster_info_ref = pool.create_actor(ClusterInfoActor, schedulers=schedulers,
                                                   service_discover_addr=service_discover_addr,
                                                   uid=ClusterInfoActor.default_name())
        # create StatusActor
        port_str = endpoint.rsplit(':', 1)[-1]
        self._status_ref = pool.create_actor(
            StatusActor, options.worker.advertise_addr + ':' + port_str,
            uid=StatusActor.default_name())

        if ignore_avail_mem:
            # start a QuotaActor instead of MemQuotaActor to avoid memory size detection
            # for debug purpose only, DON'T USE IN PRODUCTION
            self._mem_quota_ref = pool.create_actor(
                QuotaActor, options.worker.physical_memory_limit_soft, uid=MemQuotaActor.default_name())
        else:
            self._mem_quota_ref = pool.create_actor(
                MemQuotaActor, self._calc_soft_memory_limit(),
                options.worker.physical_memory_limit_hard, uid=MemQuotaActor.default_name())

        # create ChunkHolderActor
        self._chunk_holder_ref = pool.create_actor(
            ChunkHolderActor, options.worker.cache_memory_limit, uid=ChunkHolderActor.default_name())
        # create DispatchActor
        self._dispatch_ref = pool.create_actor(DispatchActor, uid=DispatchActor.default_name())
        # create ExecutionActor
        self._execution_ref = pool.create_actor(ExecutionActor, uid=ExecutionActor.default_name())

        # create CpuCalcActor
        for cpu_id in range(options.worker.cpu_process_count):
            uid = 'w:%d:mars-calc-%d-%d' % (cpu_id + 1, os.getpid(), cpu_id)
            actor = pool.create_actor(CpuCalcActor, uid=uid)
            self._cpu_calc_actors.append(actor)

        # create SenderActor and ReceiverActor
        start_pid = 1 + options.worker.cpu_process_count
        for sender_id in range(options.worker.io_process_count):
            uid = 'w:%d:mars-sender-%d-%d' % (start_pid + sender_id, os.getpid(), sender_id)
            actor = pool.create_actor(SenderActor, uid=uid)
            self._sender_actors.append(actor)
        for receiver_id in range(2 * options.worker.io_process_count):
            uid = 'w:%d:mars-receiver-%d-%d' % (start_pid + receiver_id // 2, os.getpid(), receiver_id)
            actor = pool.create_actor(ReceiverActor, uid=uid)
            self._receiver_actors.append(actor)

        # create ProcessHelperActor
        for proc_id in range(pool.cluster_info.n_process):
            uid = 'w:%d:mars-process-helper-%d-%d' % (proc_id, os.getpid(), proc_id)
            actor = pool.create_actor(ProcessHelperActor, uid=uid)
            self._process_helper_actors.append(actor)

        # create ResultSenderActor
        self._result_sender_ref = pool.create_actor(ResultSenderActor, uid=ResultSenderActor.default_name())

        # create SpillActor
        start_pid = 1 + options.worker.cpu_process_count + options.worker.io_process_count
        if options.worker.spill_directory:
            for spill_id in range(len(options.worker.spill_directory) * 2):
                uid = 'w:%d:mars-spill-%d-%d' % (start_pid, os.getpid(), spill_id)
                actor = pool.create_actor(SpillActor, uid=uid)
                self._spill_actors.append(actor)

        kv_store = kvstore.get(options.kv_store)
        if isinstance(kv_store, kvstore.EtcdKVStore):
            kv_store.write('/workers/meta_timestamp', str(int(time.time())))

    @staticmethod
    def _calc_size_limit(limit_str, total_size):
        """
        Calculate limitation size when it is represented in percentage or prettified format
        :param limit_str: percentage or prettified format
        :param total_size: total size of the container
        :return: actual size in bytes
        """
        if isinstance(limit_str, int):
            return limit_str
        mem_limit, is_percent = parse_memory_limit(limit_str)
        if is_percent:
            return int(total_size * mem_limit)
        else:
            return int(mem_limit)

    @staticmethod
    def cache_memory_limit():
        mem_stats = resource.virtual_memory()

        return WorkerService._calc_size_limit(
            options.worker.cache_memory_limit, mem_stats.total
        )

    def start_local(self, endpoint, pool, process_start_index, ignore_avail_mem=True, spill_dir=None):
        mem_stats = resource.virtual_memory()

        options.worker.physical_memory_limit_hard = self._calc_size_limit(
            options.worker.physical_memory_limit_hard, mem_stats.total
        )
        options.worker.physical_memory_limit_soft = self._calc_size_limit(
            options.worker.physical_memory_limit_soft, mem_stats.total
        )
        options.worker.cache_memory_limit = self._calc_size_limit(
            options.worker.cache_memory_limit, mem_stats.total
        )
        if spill_dir:
            from .spill import parse_spill_dirs
            spill_directory = parse_spill_dirs(spill_dir)
        else:
            spill_directory = None

        # create StatusActor
        self._status_ref = pool.create_actor(
            StatusActor, endpoint, uid=StatusActor.default_name())

        if ignore_avail_mem:
            # start a QuotaActor instead of MemQuotaActor to avoid memory size detection
            # for debug purpose only, DON'T USE IN PRODUCTION
            self._mem_quota_ref = pool.create_actor(
                QuotaActor, options.worker.physical_memory_limit_soft, uid=MemQuotaActor.default_name())
        else:
            self._mem_quota_ref = pool.create_actor(
                MemQuotaActor, self._calc_soft_memory_limit(),
                options.worker.physical_memory_limit_hard, uid=MemQuotaActor.default_name())

        # create ChunkHolderActor
        self._chunk_holder_ref = pool.create_actor(
            ChunkHolderActor, options.worker.cache_memory_limit, uid=ChunkHolderActor.default_name())
        # create DispatchActor
        self._dispatch_ref = pool.create_actor(DispatchActor, uid=DispatchActor.default_name())
        # create ExecutionActor
        self._execution_ref = pool.create_actor(ExecutionActor, uid=ExecutionActor.default_name())

        # create CpuCalcActor
        for cpu_id in range(pool.cluster_info.n_process - 1 - process_start_index):
            uid = 'w:%d:mars-calc-%d-%d' % (cpu_id + 1, os.getpid(), cpu_id)
            actor = pool.create_actor(CpuCalcActor, uid=uid)
            self._cpu_calc_actors.append(actor)

        # create ProcessHelperActor
        for proc_id in range(pool.cluster_info.n_process - process_start_index):
            uid = 'w:%d:mars-process-helper-%d-%d' % (proc_id, os.getpid(), proc_id)
            actor = pool.create_actor(ProcessHelperActor, uid=uid)
            self._process_helper_actors.append(actor)

        # create ResultSenderActor
        self._result_sender_ref = pool.create_actor(ResultSenderActor, uid=ResultSenderActor.default_name())

        # create SpillActor, put it into the last process
        start_pid = pool.cluster_info.n_process - 1
        if spill_directory:
            for spill_id in range(len(spill_directory) * 2):
                uid = 'w:%d:mars-spill-%d-%d' % (start_pid, os.getpid(), spill_id)
                actor = pool.create_actor(SpillActor, uid=uid)
                self._spill_actors.append(actor)

    def stop(self):
        if self._result_sender_ref:
            self._result_sender_ref.destroy()
        if self._status_ref:
            self._status_ref.destroy()
        if self._chunk_holder_ref:
            self._chunk_holder_ref.destroy()
        if self._dispatch_ref:
            self._dispatch_ref.destroy()
        if self._execution_ref:
            self._execution_ref.destroy()

        for actor in (self._cpu_calc_actors + self._sender_actors
                      + self._receiver_actors + self._spill_actors + self._process_helper_actors):
            actor.destroy()

        self._plasma_store.__exit__(None, None, None)



# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from collections import deque

from .utils import WorkerActor
from ..utils import mod_hash, log_unhandled

logger = logging.getLogger(__name__)


class DispatchActor(WorkerActor):
    """
    Dispatcher for multiple actors belonging to one category
    """
    def __init__(self):
        super(DispatchActor, self).__init__()
        self._free_slots = dict()
        self._all_slots = dict()
        self._free_slot_requests = dict()

    def post_create(self):
        super(DispatchActor, self).post_create()

    @log_unhandled
    def get_free_slot(self, queue_name, callback=None):
        """
        Get uid of a free actor when available
        :param queue_name: queue name
        :param callback: promise callback
        """
        if queue_name not in self._free_slots:
            self.tell_promise(callback, None)
            return
        if not self._free_slots[queue_name]:
            # no slots free, we queue the callback
            self._free_slot_requests[queue_name].append(callback)
            logger.debug('No valid slots available. slot dump: %r', self._dump_free_slots())
            return
        self.tell_promise(callback, self._free_slots[queue_name].pop())

    @log_unhandled
    def get_hash_slot(self, queue_name, key):
        """
        Get uid of a slot by hash value
        :param queue_name: queue name
        :param key: key to be hashed
        """
        slots = self._all_slots[queue_name]
        uid = slots[mod_hash(key, len(slots))]
        return uid

    @log_unhandled
    def get_slots(self, queue_name):
        """
        Get all uids of slots of a queue
        :param queue_name: queue name
        """
        if queue_name not in self._all_slots:
            return []
        return list(self._all_slots[queue_name])

    def _dump_free_slots(self):
        return dict((k, len(v)) for k, v in self._free_slots.items())

    @log_unhandled
    def register_free_slot(self, uid, queue_name):
        """
        Register free uid of a queue
        :param uid: uid of free actor
        :param queue_name: queue name
        """
        if queue_name not in self._free_slots:
            self._free_slot_requests[queue_name] = deque()
            self._free_slots[queue_name] = []
            self._all_slots[queue_name] = []
        self._free_slots[queue_name].append(uid)
        self._all_slots[queue_name].append(uid)

        if self._free_slot_requests[queue_name]:
            self.tell_promise(self._free_slot_requests[queue_name].popleft(),
                              self._free_slots[queue_name].pop())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import weakref

from ..compat import six
from ..serialize import SerializableMetaclass, AttributeAsDictKey, ValueType, ProviderType, \
    IdentityField, ListField, DataTypeField, Int32Field, BoolField, DictField
from ..core import Entity
from ..utils import AttributeDict


operand_type_to_oprand_cls = {}
OP_TYPE_KEY = '_op_type_'
OP_MODULE_KEY = '_op_module_'


class OperandMetaclass(SerializableMetaclass):
    def __new__(mcs, name, bases, kv):
        cls = super(OperandMetaclass, mcs).__new__(mcs, name, bases, kv)

        for base in bases:
            if OP_TYPE_KEY not in kv and hasattr(base, OP_TYPE_KEY):
                kv[OP_TYPE_KEY] = getattr(base, OP_TYPE_KEY)
            if OP_MODULE_KEY not in kv and hasattr(base, OP_MODULE_KEY):
                kv[OP_MODULE_KEY] = getattr(base, OP_MODULE_KEY)

        if kv.get(OP_TYPE_KEY) is not None and kv.get(OP_MODULE_KEY) is not None:
            # common operand can be inherited for different modules, like tensor or dataframe, so forth
            operand_type_to_oprand_cls[kv[OP_MODULE_KEY], kv[OP_TYPE_KEY]] = cls

        return cls


class Operand(six.with_metaclass(OperandMetaclass, AttributeAsDictKey)):
    """
    Operand base class. All operands should have a type, which can be Add, Subtract etc.
    `sparse` indicates that if the operand is applied on a sparse tensor/chunk.
    `gpu` indicates that if the operand should be executed on the GPU.
    `device`, 0 means the CPU, otherwise means the GPU device.
    Operand can have inputs and outputs
    which should be the :class:`mars.tensor.core.TensorData`, :class:`mars.tensor.core.ChunkData` etc.
    """

    attr_tag = 'attr'
    _init_update_key_ = False

    _op_id = IdentityField('type')

    _sparse = BoolField('sparse')
    _gpu = BoolField('gpu')
    _device = Int32Field('device')

    _dtype = DataTypeField('dtype')

    _inputs = ListField('inputs', ValueType.key)
    _outputs = ListField('outputs', ValueType.key, weak_ref=True)

    _params = DictField('params', key_type=ValueType.string, on_deserialize=AttributeDict)

    def __new__(cls, *args, **kwargs):
        if '_op_id' in kwargs and kwargs['_op_id']:
            op_id = kwargs['_op_id']
            module, tp = op_id.split('.', 1)
            cls = operand_type_to_oprand_cls[module, int(tp)]
        return object.__new__(cls)

    def __init__(self, *args, **kwargs):
        extras = AttributeDict((k, kwargs.pop(k)) for k in set(kwargs) - set(self.__slots__))
        kwargs['_params'] = kwargs.pop('_params', extras)
        super(Operand, self).__init__(*args, **kwargs)
        if hasattr(self, OP_MODULE_KEY) and hasattr(self, OP_TYPE_KEY):
            self._op_id = '{0}.{1}'.format(getattr(self, OP_MODULE_KEY), getattr(self, OP_TYPE_KEY))

    @classmethod
    def cls(cls, provider):
        if provider.type == ProviderType.protobuf:
            from ..serialize.protos.operand_pb2 import OperandDef
            return OperandDef
        return super(Operand, cls).cls(provider)

    @property
    def inputs(self):
        return getattr(self, '_inputs', None)

    @inputs.setter
    def inputs(self, vals):
        self._set_inputs(vals)

    @property
    def outputs(self):
        outputs = getattr(self, '_outputs', None)
        if outputs:
            return [ref() for ref in outputs]

    @outputs.setter
    def outputs(self, outputs):
        self._attach_outputs(*outputs)

    @property
    def output_limit(self):
        return 1

    @property
    def dtype(self):
        return getattr(self, '_dtype', None)

    @property
    def gpu(self):
        return getattr(self, '_gpu', False)

    @property
    def device(self):
        return getattr(self, '_device', None)

    @property
    def params(self):
        return self._params

    @params.setter
    def params(self, params):
        self._params = params

    @property
    def sparse(self):
        return getattr(self, '_sparse', False)

    def is_sparse(self):
        return getattr(self, '_sparse', False) or False

    issparse = is_sparse

    def is_gpu(self):
        return getattr(self, '_gpu', False) or False

    @classmethod
    def _get_entity_data(cls, entity):
        if isinstance(entity, Entity):
            return entity.data
        return entity

    @classmethod
    def _get_inputs_data(cls, inputs):
        return [cls._get_entity_data(inp) for inp in inputs]

    def _set_inputs(self, inputs):
        if inputs is not None:
            inputs = self._get_inputs_data(inputs)
        if hasattr(self, 'check_inputs'):
            self.check_inputs(inputs)
        setattr(self, '_inputs', inputs)

    def _attach_outputs(self, *outputs):
        self._outputs = tuple(weakref.ref(self._get_entity_data(o)) if o is not None else o
                              for o in outputs)

        if len(self._outputs) > self.output_limit:
            raise ValueError("Outputs' size exceeds limitation")

    def copy(self):
        new_op = super(Operand, self).copy()
        new_op.outputs = []

        return new_op



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from collections import Iterable
from contextlib import contextmanager

import numpy as np

from ....config import options
from ....compat import irange, izip
from ....operands.random import State
from ...core import TENSOR_TYPE, CHUNK_TYPE
from ..utils import decide_chunks, random_state_data, broadcast_shape
from ..core import TensorOperandMixin
from ..datasource import tensor as astensor
from ..base import broadcast_to


class RandomState(object):
    def __init__(self, seed=None):
        self._random_state = np.random.RandomState(seed=seed)
        self._curr_seed = seed

    def seed(self, seed=None):
        """
        Seed the generator.

        This method is called when `RandomState` is initialized. It can be
        called again to re-seed the generator. For details, see `RandomState`.

        Parameters
        ----------
        seed : int or 1-d array_like, optional
            Seed for `RandomState`.
            Must be convertible to 32 bit unsigned integers.

        See Also
        --------
        RandomState
        """
        self._random_state.seed(seed=seed)
        self._curr_seed = seed

    @property
    def _state(self):
        return State(self._random_state) if self._curr_seed is not None else None

    @classmethod
    def _handle_size(cls, size):
        if size is None:
            return size
        try:
            return tuple(size)
        except TypeError:
            return size,


_random_state = RandomState()


def handle_array(arg):
    if not isinstance(arg, TENSOR_TYPE):
        if not isinstance(arg, Iterable):
            return arg

        arg = np.asarray(arg)
        return arg[(0,) * max(1, arg.ndim)]
    elif hasattr(arg, 'op') and hasattr(arg.op, 'data'):
        return arg.op.data[(0,) * max(1, arg.ndim)]

    return np.empty((0,), dtype=arg.dtype)


class TensorRandomOperandMixin(TensorOperandMixin):
    __slots__ = ()

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]
        chunks = tensor.params.raw_chunks or options.tensor.chunks
        nsplits = decide_chunks(tensor.shape, chunks, tensor.dtype.itemsize)
        fields = getattr(op, '_input_fields_', [])
        to_one_chunk_fields = set(getattr(op, '_into_one_chunk_fields_', list()))

        new_inputs = []
        changed = False
        for field in fields:
            t = getattr(op, field)
            if not isinstance(t, TENSOR_TYPE):
                continue

            if field not in to_one_chunk_fields:
                t_nsplits = nsplits
            else:
                t_nsplits = t.shape  # into 1 chunk
            rechunked = t.rechunk(t_nsplits)
            if rechunked is not t:
                rechunked.single_tiles()
                changed = True
                new_inputs.append(rechunked)
            else:
                new_inputs.append(t)
        if changed:
            op.inputs = new_inputs

        idxes = list(itertools.product(*[irange(len(s)) for s in nsplits]))
        states = random_state_data(len(idxes), op.state.random_state) \
            if op.state is not None else [None] * len(idxes)

        out_chunks = []
        for state, idx, shape in izip(states, idxes, itertools.product(*nsplits)):
            inputs = []
            for inp in op.inputs:
                if len(inp.chunks) == 1:
                    inputs.append(inp.chunks[0])
                else:
                    inputs.append(inp.cix[idx])
            state = State(np.random.RandomState(state)) \
                if state is not None else None
            try:
                s = len(tuple(op.size))
                size = shape[:s]
            except TypeError:
                if op.size is None:
                    size = None
                else:
                    size = shape[:1]
            except AttributeError:
                size = shape

            chunk_op = op.copy().reset_key()
            chunk_op._size = size
            chunk_op._state = state
            out_chunk = chunk_op.new_chunk(inputs, shape, index=idx)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, tensor.shape,
                                  chunks=out_chunks, nsplits=nsplits)

    def _get_shape(self, shapes):
        shapes = list(shapes)
        if getattr(self, '_size', None) is not None:
            shapes.append(getattr(self, '_size'))
        return broadcast_shape(*shapes)

    @classmethod
    def _handle_arg(cls, arg, chunks):
        if isinstance(arg, (list, np.ndarray)):
            arg = astensor(arg, chunks=chunks)

        return arg

    @contextmanager
    def _get_inputs_shape_by_given_fields(self, inputs, shape, raw_chunks=None, tensor=True):
        fields = getattr(self, '_input_fields_', [])
        to_one_chunk_fields = set(getattr(self, '_into_one_chunk_fields_', list()))

        field_to_obj = dict()
        to_broadcast_shapes = []
        if fields:
            if getattr(self, fields[0], None) is None:
                # create from beginning
                for i, field, val in zip(itertools.count(0), fields, inputs):
                    if field not in to_one_chunk_fields:
                        if isinstance(val, list):
                            val = np.asarray(val)
                        if tensor:
                            val = self._handle_arg(val, raw_chunks)
                    if isinstance(val, TENSOR_TYPE + CHUNK_TYPE):
                        field_to_obj[field] = val
                        if field not in to_one_chunk_fields:
                            to_broadcast_shapes.append(val.shape)
                    setattr(self, field, val)
            else:
                inputs_iter = iter(inputs)
                for i, field in enumerate(fields):
                    if isinstance(getattr(self, field), TENSOR_TYPE + CHUNK_TYPE):
                        field_to_obj[field] = next(inputs_iter)

        if tensor:
            if shape is None:
                shape = self._get_shape(to_broadcast_shapes)

            for field, inp in field_to_obj.items():
                if field not in to_one_chunk_fields:
                    field_to_obj[field] = broadcast_to(inp, shape)

        yield [field_to_obj[f] for f in fields if f in field_to_obj], shape

        inputs_iter = iter(getattr(self, '_inputs'))
        for field in fields:
            if field in field_to_obj:
                setattr(self, field, next(inputs_iter))

    def new_tensors(self, inputs, shape, **kw):
        raw_chunks = kw.get('chunks', None)
        with self._get_inputs_shape_by_given_fields(inputs, shape, raw_chunks, True) as (inputs, shape):
            return super(TensorRandomOperandMixin, self).new_tensors(inputs, shape, **kw)

    def new_chunks(self, inputs, shape, **kw):
        with self._get_inputs_shape_by_given_fields(inputs, shape, None, False) as (inputs, shape):
            return super(TensorRandomOperandMixin, self).new_chunks(inputs, shape, **kw)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin, handle_array


class TensorWeibull(operands.Weibull, TensorRandomOperandMixin):
    __slots__ = '_a', '_size'
    _input_fields_ = ['_a']

    def __init__(self, size=None, state=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorWeibull, self).__init__(_size=size, _state=state, _dtype=dtype,
                                            _gpu=gpu, **kw)

    def __call__(self, a, chunks=None):
        return self.new_tensor([a], None, raw_chunks=chunks)


def weibull(random_state, a, size=None, chunks=None, gpu=None, **kw):
    """
    Draw samples from a Weibull distribution.

    Draw samples from a 1-parameter Weibull distribution with the given
    shape parameter `a`.

    .. math:: X = (-ln(U))^{1/a}

    Here, U is drawn from the uniform distribution over (0,1].

    The more common 2-parameter Weibull, including a scale parameter
    :math:`\lambda` is just :math:`X = \lambda(-ln(U))^{1/a}`.

    Parameters
    ----------
    a : float or array_like of floats
        Shape of the distribution. Should be greater than zero.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``a`` is a scalar.  Otherwise,
        ``mt.array(a).size`` samples are drawn.
    chunks : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized Weibull distribution.

    See Also
    --------
    scipy.stats.weibull_max
    scipy.stats.weibull_min
    scipy.stats.genextreme
    gumbel

    Notes
    -----
    The Weibull (or Type III asymptotic extreme value distribution
    for smallest values, SEV Type III, or Rosin-Rammler
    distribution) is one of a class of Generalized Extreme Value
    (GEV) distributions used in modeling extreme value problems.
    This class includes the Gumbel and Frechet distributions.

    The probability density for the Weibull distribution is

    .. math:: p(x) = \frac{a}
                     {\lambda}(\frac{x}{\lambda})^{a-1}e^{-(x/\lambda)^a},

    where :math:`a` is the shape and :math:`\lambda` the scale.

    The function has its peak (the mode) at
    :math:`\lambda(\frac{a-1}{a})^{1/a}`.

    When ``a = 1``, the Weibull distribution reduces to the exponential
    distribution.

    References
    ----------
    .. [1] Waloddi Weibull, Royal Technical University, Stockholm,
           1939 "A Statistical Theory Of The Strength Of Materials",
           Ingeniorsvetenskapsakademiens Handlingar Nr 151, 1939,
           Generalstabens Litografiska Anstalts Forlag, Stockholm.
    .. [2] Waloddi Weibull, "A Statistical Distribution Function of
           Wide Applicability", Journal Of Applied Mechanics ASME Paper
           1951.
    .. [3] Wikipedia, "Weibull distribution",
           http://en.wikipedia.org/wiki/Weibull_distribution

    Examples
    --------
    Draw samples from the distribution:

    >>> import mars.tensor as mt

    >>> a = 5. # shape
    >>> s = mt.random.weibull(a, 1000)

    Display the histogram of the samples, along with
    the probability density function:

    >>> import matplotlib.pyplot as plt
    >>> x = mt.arange(1,100.)/50.
    >>> def weib(x,n,a):
    ...     return (a / n) * (x / n)**(a - 1) * mt.exp(-(x / n)**a)

    >>> count, bins, ignored = plt.hist(mt.random.weibull(5.,1000).execute())
    >>> x = mt.arange(1,100.)/50.
    >>> scale = count.max()/weib(x, 1., 5.).max()
    >>> plt.plot(x.execute(), (weib(x, 1., 5.)*scale).execute())
    >>> plt.show()
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().weibull(
            handle_array(a), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorWeibull(size=size, state=random_state._state, gpu=gpu, **kw)
    return op(a, chunks=chunks)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from string import ascii_letters

from ..expressions import arithmetic
from ..expressions.fuse import TensorCpFuseChunk
from ...utils import tokenize


CP_BINOP_TO_STRING = {
    arithmetic.TensorSubtract: '-',
    arithmetic.TensorSubConstant: '-',
    arithmetic.TensorMultiply: '*',
    arithmetic.TensorMulConstant: '*',
    arithmetic.TensorTrueDiv: '/',
}

CP_UNARYOP_TO_STRING = {
    arithmetic.TensorSqrt: 'sqrt',
}


def _evaluate(chunk):
    letters = iter(l for l in ascii_letters if l not in 'ni')

    input_types = [i.dtype.name for i in chunk.op.inputs]
    input_names = {i: next(letters) for i in chunk.op.inputs}
    input_arguments = ', '.join(['{0} {1}'.format(tp, input_names[i])
                                 for i, tp in zip(chunk.op.inputs, input_types)])
    output_type = chunk.op.dtype.name
    output_name = next(letters)
    output_argument = '{0} {1}'.format(output_type, output_name)
    body = dict(input_names)

    for node in chunk.composed:
        if type(node.op) in CP_BINOP_TO_STRING:
            input_bodies = [body.get(i, repr(i)) for i in (node.op.lhs, node.op.rhs)]
            body[node] = ' {0} '.format(CP_BINOP_TO_STRING[type(node.op)]).join(input_bodies)
        elif type(node.op) in CP_UNARYOP_TO_STRING:
            body[node] = '{0}({1})'.format(CP_UNARYOP_TO_STRING[type(node.op)], body[node.op.inputs[0]])
        else:
            raise NotImplementedError

    body = '{0} = {1}'.format(output_name, body[chunk.composed[-1]])
    return input_arguments, output_argument, body, \
           '{0}_{1}'.format(chunk.op.__class__.__name__.lower(),
                            tokenize(input_arguments, output_argument, body))


def _execute_cp(ctx, chunk):
    import cupy as cp

    func = cp.ElementwiseKernel(*_evaluate(chunk))
    ctx[chunk.key] = func(*[ctx[i.key] for i in chunk.op.inputs])


def register_cp_handler():
    from .core import register

    register(TensorCpFuseChunk, _execute_cp)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .... import operands
from ..datasource import tensor as astensor
from .core import TensorReduction


class TensorNanMax(operands.NanMax, TensorReduction):
    def __init__(self, axis=None, dtype=None, keepdims=None, combine_size=None, **kw):
        super(TensorNanMax, self).__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                                           _combine_size=combine_size, **kw)

    @staticmethod
    def _get_op_types():
        return TensorNanMax, TensorNanMax, None


def nanmax(a, axis=None, out=None, keepdims=None, combine_size=None):
    """
    Return the maximum of an array or maximum along an axis, ignoring any
    NaNs.  When all-NaN slices are encountered a ``RuntimeWarning`` is
    raised and NaN is returned for that slice.

    Parameters
    ----------
    a : array_like
        Tensor containing numbers whose maximum is desired. If `a` is not a
        tensor, a conversion is attempted.
    axis : int, optional
        Axis along which the maximum is computed. The default is to compute
        the maximum of the flattened tensor.
    out : ndarray, optional
        Alternate output array in which to place the result.  The default
        is ``None``; if provided, it must have the same shape as the
        expected output, but the type will be cast if necessary.  See
        `doc.ufuncs` for details.
    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one. With this option,
        the result will broadcast correctly against the original `a`.

        If the value is anything but the default, then
        `keepdims` will be passed through to the `max` method
        of sub-classes of `Tensor`.  If the sub-classes methods
        does not implement `keepdims` any exceptions will be raised.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    nanmax : Tensor
        A tensor with the same shape as `a`, with the specified axis removed.
        If `a` is a 0-d tensor, or if axis is None, a Tensor scalar is
        returned.  The same dtype as `a` is returned.

    See Also
    --------
    nanmin :
        The minimum value of a tensor along a given axis, ignoring any NaNs.
    amax :
        The maximum value of a tensor along a given axis, propagating any NaNs.
    fmax :
        Element-wise maximum of two tensors, ignoring any NaNs.
    maximum :
        Element-wise maximum of two tensors, propagating any NaNs.
    isnan :
        Shows which elements are Not a Number (NaN).
    isfinite:
        Shows which elements are neither NaN nor infinity.

    amin, fmin, minimum

    Notes
    -----
    Mars uses the IEEE Standard for Binary Floating-Point for Arithmetic
    (IEEE 754). This means that Not a Number is not equivalent to infinity.
    Positive infinity is treated as a very large number and negative
    infinity is treated as a very small (i.e. negative) number.

    If the input has a integer type the function is equivalent to np.max.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array([[1, 2], [3, mt.nan]])
    >>> mt.nanmax(a).execute()
    3.0
    >>> mt.nanmax(a, axis=0).execute()
    array([ 3.,  2.])
    >>> mt.nanmax(a, axis=1).execute()
    array([ 2.,  3.])

    When positive infinity and negative infinity are present:

    >>> mt.nanmax([1, 2, mt.nan, mt.NINF]).execute()
    2.0
    >>> mt.nanmax([1, 2, mt.nan, mt.inf]).execute()
    inf

    """
    a = astensor(a)
    op = TensorNanMax(axis=axis, dtype=a.dtype, keepdims=keepdims, combine_size=combine_size)
    return op(a, out=out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from ..utils import infer_dtype
from .core import TensorBinOp, TensorConstant


class TensorBitand(operands.Bitand, TensorBinOp):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorBitand, self).__init__(_casting=casting, _err=err,
                                           _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        return x1.issparse() or x2.issparse()

    @classmethod
    def constant_cls(cls):
        return TensorBitandConstant


class TensorBitandConstant(operands.BitandConstant, TensorConstant):
    def __init__(self, casting='same_kind', err=None, dtype=None, sparse=False, **kw):
        err = err if err is not None else np.geterr()
        super(TensorBitandConstant, self).__init__(_casting=casting, _err=err,
                                                   _dtype=dtype, _sparse=sparse, **kw)

    @classmethod
    def _is_sparse(cls, x1, x2):
        if (hasattr(x1, 'issparse') and x1.issparse()) or \
                (hasattr(x2, 'issparse') and x2.issparse()):
            return True
        return False


@infer_dtype(np.bitwise_and)
def bitand(x1, x2, out=None, where=None, **kwargs):
    """
    Compute the bit-wise AND of two tensors element-wise.

    Computes the bit-wise AND of the underlying binary representation of
    the integers in the input arrays. This ufunc implements the C/Python
    operator ``&``.

    Parameters
    ----------
    x1, x2 : array_like
        Only integer and boolean types are handled.
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    out : array_like
        Result.

    See Also
    --------
    logical_and
    bitwise_or
    bitwise_xor

    Examples
    --------
    The number 13 is represented by ``00001101``.  Likewise, 17 is
    represented by ``00010001``.  The bit-wise AND of 13 and 17 is
    therefore ``000000001``, or 1:

    >>> import mars.tensor as mt

    >>> mt.bitwise_and(13, 17).execute()
    1

    >>> mt.bitwise_and(14, 13).execute()
    12
    >>> mt.bitwise_and([14,3], 13).execute()
    array([12,  1])

    >>> mt.bitwise_and([11,7], [4,25]).execute()
    array([0, 1])
    >>> mt.bitwise_and(mt.array([2,5,255]), mt.array([3,14,16])).execute()
    array([ 2,  4, 16])
    >>> mt.bitwise_and([True, True], [False, True]).execute()
    array([False,  True])
    """
    op = TensorBitand(**kwargs)
    return op(x1, x2, out=out, where=where)


@infer_dtype(np.bitwise_and, reverse=True)
def rbitand(x1, x2, **kwargs):
    op = TensorBitand(**kwargs)
    return op.rcall(x1, x2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ....operands import fft as fftop
from ....compat import izip
from ..datasource import tensor as astensor
from .core import TensorFFTNMixin, validate_fftn


class TensorRFFT2(fftop.RFFT2, TensorFFTNMixin):
    def __init__(self, shape=None, axes=None, norm=None, dtype=None, **kw):
        super(TensorRFFT2, self).__init__(_shape=shape, _axes=axes, _norm=norm,
                                          _dtype=dtype, **kw)

    @classmethod
    def _get_shape(cls, op, shape):
        new_shape = list(shape)
        if op.shape is not None:
            for ss, axis in izip(op.shape, op.axes):
                new_shape[axis] = ss
        new_shape[op.axes[-1]] = new_shape[op.axes[-1]] // 2 + 1
        return tuple(new_shape)

    def _set_inputs(self, inputs):
        super(TensorRFFT2, self)._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, a):
        shape = self._get_shape(self, a.shape)
        return self.new_tensor([a], shape)


def rfft2(a, s=None, axes=(-2, -1), norm=None):
    """
    Compute the 2-dimensional FFT of a real tensor.

    Parameters
    ----------
    a : array_like
        Input tensor, taken to be real.
    s : sequence of ints, optional
        Shape of the FFT.
    axes : sequence of ints, optional
        Axes over which to compute the FFT.
    norm : {None, "ortho"}, optional
        Normalization mode (see `mt.fft`). Default is None.

    Returns
    -------
    out : Tensor
        The result of the real 2-D FFT.

    See Also
    --------
    rfftn : Compute the N-dimensional discrete Fourier Transform for real
            input.

    Notes
    -----
    This is really just `rfftn` with different default behavior.
    For more details see `rfftn`.

    """
    if len(axes) != 2:
        raise ValueError("axes length should be 2")
    a = astensor(a)
    axes = validate_fftn(a, s=s, axes=axes, norm=norm)
    op = TensorRFFT2(shape=s, axes=axes, norm=norm, dtype=np.dtype(np.complex_))
    return op(a)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin, handle_array


class TensorTriangular(operands.Triangular, TensorRandomOperandMixin):
    __slots__ = '_left', '_mode', '_right', '_size'
    _input_fields_ = ['_left', '_mode', '_right']

    def __init__(self, size=None, state=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorTriangular, self).__init__(_size=size, _state=state, _dtype=dtype,
                                               _gpu=gpu, **kw)

    def __call__(self, left, mode, right, chunks=None):
        return self.new_tensor([left, mode, right], None, raw_chunks=chunks)


def triangular(random_state, left, mode, right, size=None, chunks=None, gpu=None, **kw):
    """
    Draw samples from the triangular distribution over the
    interval ``[left, right]``.

    The triangular distribution is a continuous probability
    distribution with lower limit left, peak at mode, and upper
    limit right. Unlike the other distributions, these parameters
    directly define the shape of the pdf.

    Parameters
    ----------
    left : float or array_like of floats
        Lower limit.
    mode : float or array_like of floats
        The value where the peak of the distribution occurs.
        The value should fulfill the condition ``left <= mode <= right``.
    right : float or array_like of floats
        Upper limit, should be larger than `left`.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``left``, ``mode``, and ``right``
        are all scalars.  Otherwise, ``mt.broadcast(left, mode, right).size``
        samples are drawn.
    chunks : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized triangular distribution.

    Notes
    -----
    The probability density function for the triangular distribution is

    .. math:: P(x;l, m, r) = \begin{cases}
              \frac{2(x-l)}{(r-l)(m-l)}& \text{for $l \leq x \leq m$},\\
              \frac{2(r-x)}{(r-l)(r-m)}& \text{for $m \leq x \leq r$},\\
              0& \text{otherwise}.
              \end{cases}

    The triangular distribution is often used in ill-defined
    problems where the underlying distribution is not known, but
    some knowledge of the limits and mode exists. Often it is used
    in simulations.

    References
    ----------
    .. [1] Wikipedia, "Triangular distribution"
           http://en.wikipedia.org/wiki/Triangular_distribution

    Examples
    --------
    Draw values from the distribution and plot the histogram:

    >>> import matplotlib.pyplot as plt
    >>> import mars.tensor as mt
    >>> h = plt.hist(mt.random.triangular(-3, 0, 8, 100000).execute(), bins=200,
    ...              normed=True)
    >>> plt.show()
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().triangular(
            handle_array(left), handle_array(mode), handle_array(right), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorTriangular(size=size, state=random_state._state, gpu=gpu, **kw)
    return op(left, mode, right, chunks=chunks)



"""functools.py - Tools for working with functions and callable objects
"""
# Python module wrapper for _functools C module
# to allow utilities written in Python to be added
# to the functools module.
# Written by Nick Coghlan <ncoghlan at gmail.com>
# and Raymond Hettinger <python at rcn.com>
#   Copyright (C) 2006-2010 Python Software Foundation.
# See C source code for _functools credits/copyright

__all__ = ['update_wrapper', 'wraps', 'WRAPPER_ASSIGNMENTS', 'WRAPPER_UPDATES',
           'total_ordering', 'cmp_to_key', 'lru_cache', 'reduce', 'partial']

from _functools import partial, reduce
from collections import MutableMapping, namedtuple
from .reprlib32 import recursive_repr as _recursive_repr
from weakref import proxy as _proxy
import sys as _sys
try:
    from thread import allocate_lock as Lock
except ImportError:
    from ._dummy_thread32 import allocate_lock as Lock

################################################################################
### OrderedDict
################################################################################

class _Link(object):
    __slots__ = 'prev', 'next', 'key', '__weakref__'

class OrderedDict(dict):
    'Dictionary that remembers insertion order'
    # An inherited dict maps keys to values.
    # The inherited dict provides __getitem__, __len__, __contains__, and get.
    # The remaining methods are order-aware.
    # Big-O running times for all methods are the same as regular dictionaries.

    # The internal self.__map dict maps keys to links in a doubly linked list.
    # The circular doubly linked list starts and ends with a sentinel element.
    # The sentinel element never gets deleted (this simplifies the algorithm).
    # The sentinel is in self.__hardroot with a weakref proxy in self.__root.
    # The prev links are weakref proxies (to prevent circular references).
    # Individual links are kept alive by the hard reference in self.__map.
    # Those hard references disappear when a key is deleted from an OrderedDict.

    def __init__(self, *args, **kwds):
        '''Initialize an ordered dictionary.  The signature is the same as
        regular dictionaries, but keyword arguments are not recommended because
        their insertion order is arbitrary.

        '''
        if len(args) > 1:
            raise TypeError('expected at most 1 arguments, got %d' % len(args))
        try:
            self.__root
        except AttributeError:
            self.__hardroot = _Link()
            self.__root = root = _proxy(self.__hardroot)
            root.prev = root.next = root
            self.__map = {}
        self.__update(*args, **kwds)

    def __setitem__(self, key, value,
                    dict_setitem=dict.__setitem__, proxy=_proxy, Link=_Link):
        'od.__setitem__(i, y) <==> od[i]=y'
        # Setting a new item creates a new link at the end of the linked list,
        # and the inherited dictionary is updated with the new key/value pair.
        if key not in self:
            self.__map[key] = link = Link()
            root = self.__root
            last = root.prev
            link.prev, link.next, link.key = last, root, key
            last.next = link
            root.prev = proxy(link)
        dict_setitem(self, key, value)

    def __delitem__(self, key, dict_delitem=dict.__delitem__):
        'od.__delitem__(y) <==> del od[y]'
        # Deleting an existing item uses self.__map to find the link which gets
        # removed by updating the links in the predecessor and successor nodes.
        dict_delitem(self, key)
        link = self.__map.pop(key)
        link_prev = link.prev
        link_next = link.next
        link_prev.next = link_next
        link_next.prev = link_prev

    def __iter__(self):
        'od.__iter__() <==> iter(od)'
        # Traverse the linked list in order.
        root = self.__root
        curr = root.next
        while curr is not root:
            yield curr.key
            curr = curr.next

    def __reversed__(self):
        'od.__reversed__() <==> reversed(od)'
        # Traverse the linked list in reverse order.
        root = self.__root
        curr = root.prev
        while curr is not root:
            yield curr.key
            curr = curr.prev

    def clear(self):
        'od.clear() -> None.  Remove all items from od.'
        root = self.__root
        root.prev = root.next = root
        self.__map.clear()
        dict.clear(self)

    def popitem(self, last=True):
        '''od.popitem() -> (k, v), return and remove a (key, value) pair.
        Pairs are returned in LIFO order if last is true or FIFO order if false.

        '''
        if not self:
            raise KeyError('dictionary is empty')
        root = self.__root
        if last:
            link = root.prev
            link_prev = link.prev
            link_prev.next = root
            root.prev = link_prev
        else:
            link = root.next
            link_next = link.next
            root.next = link_next
            link_next.prev = root
        key = link.key
        del self.__map[key]
        value = dict.pop(self, key)
        return key, value

    def move_to_end(self, key, last=True):
        '''Move an existing element to the end (or beginning if last==False).

        Raises KeyError if the element does not exist.
        When last=True, acts like a fast version of self[key]=self.pop(key).

        '''
        link = self.__map[key]
        link_prev = link.prev
        link_next = link.next
        link_prev.next = link_next
        link_next.prev = link_prev
        root = self.__root
        if last:
            last = root.prev
            link.prev = last
            link.next = root
            last.next = root.prev = link
        else:
            first = root.next
            link.prev = root
            link.next = first
            root.next = first.prev = link

    def __sizeof__(self):
        sizeof = _sys.getsizeof
        n = len(self) + 1                       # number of links including root
        size = sizeof(self.__dict__)            # instance dictionary
        size += sizeof(self.__map) * 2          # internal dict and inherited dict
        size += sizeof(self.__hardroot) * n     # link objects
        size += sizeof(self.__root) * n         # proxy objects
        return size

    update = __update = MutableMapping.update
    keys = MutableMapping.keys
    values = MutableMapping.values
    items = MutableMapping.items
    __ne__ = MutableMapping.__ne__

    __marker = object()

    def pop(self, key, default=__marker):
        '''od.pop(k[,d]) -> v, remove specified key and return the corresponding
        value.  If key is not found, d is returned if given, otherwise KeyError
        is raised.

        '''
        if key in self:
            result = self[key]
            del self[key]
            return result
        if default is self.__marker:
            raise KeyError(key)
        return default

    def setdefault(self, key, default=None):
        'od.setdefault(k[,d]) -> od.get(k,d), also set od[k]=d if k not in od'
        if key in self:
            return self[key]
        self[key] = default
        return default

    @_recursive_repr()
    def __repr__(self):
        'od.__repr__() <==> repr(od)'
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self.items()))

    def __reduce__(self):
        'Return state information for pickling'
        items = [[k, self[k]] for k in self]
        inst_dict = vars(self).copy()
        for k in vars(OrderedDict()):
            inst_dict.pop(k, None)
        if inst_dict:
            return (self.__class__, (items,), inst_dict)
        return self.__class__, (items,)

    def copy(self):
        'od.copy() -> a shallow copy of od'
        return self.__class__(self)

    @classmethod
    def fromkeys(cls, iterable, value=None):
        '''OD.fromkeys(S[, v]) -> New ordered dictionary with keys from S.
        If not specified, the value defaults to None.

        '''
        self = cls()
        for key in iterable:
            self[key] = value
        return self

    def __eq__(self, other):
        '''od.__eq__(y) <==> od==y.  Comparison to another OD is order-sensitive
        while comparison to a regular mapping is order-insensitive.

        '''
        if isinstance(other, OrderedDict):
            return len(self)==len(other) and \
                   all(p==q for p, q in zip(self.items(), other.items()))
        return dict.__eq__(self, other)

# update_wrapper() and wraps() are tools to help write
# wrapper functions that can handle naive introspection

WRAPPER_ASSIGNMENTS = ('__module__', '__name__', '__doc__')
WRAPPER_UPDATES = ('__dict__',)
def update_wrapper(wrapper,
                   wrapped,
                   assigned = WRAPPER_ASSIGNMENTS,
                   updated = WRAPPER_UPDATES):
    """Update a wrapper function to look like the wrapped function

       wrapper is the function to be updated
       wrapped is the original function
       assigned is a tuple naming the attributes assigned directly
       from the wrapped function to the wrapper function (defaults to
       functools.WRAPPER_ASSIGNMENTS)
       updated is a tuple naming the attributes of the wrapper that
       are updated with the corresponding attribute from the wrapped
       function (defaults to functools.WRAPPER_UPDATES)
    """
    wrapper.__wrapped__ = wrapped
    for attr in assigned:
        try:
            value = getattr(wrapped, attr)
        except AttributeError:
            pass
        else:
            setattr(wrapper, attr, value)
    for attr in updated:
        getattr(wrapper, attr).update(getattr(wrapped, attr, {}))
    # Return the wrapper so this can be used as a decorator via partial()
    return wrapper

def wraps(wrapped,
          assigned = WRAPPER_ASSIGNMENTS,
          updated = WRAPPER_UPDATES):
    """Decorator factory to apply update_wrapper() to a wrapper function

       Returns a decorator that invokes update_wrapper() with the decorated
       function as the wrapper argument and the arguments to wraps() as the
       remaining arguments. Default arguments are as for update_wrapper().
       This is a convenience function to simplify applying partial() to
       update_wrapper().
    """
    return partial(update_wrapper, wrapped=wrapped,
                   assigned=assigned, updated=updated)

def total_ordering(cls):
    """Class decorator that fills in missing ordering methods"""
    convert = {
        '__lt__': [('__gt__', lambda self, other: not (self < other or self == other)),
                   ('__le__', lambda self, other: self < other or self == other),
                   ('__ge__', lambda self, other: not self < other)],
        '__le__': [('__ge__', lambda self, other: not self <= other or self == other),
                   ('__lt__', lambda self, other: self <= other and not self == other),
                   ('__gt__', lambda self, other: not self <= other)],
        '__gt__': [('__lt__', lambda self, other: not (self > other or self == other)),
                   ('__ge__', lambda self, other: self > other or self == other),
                   ('__le__', lambda self, other: not self > other)],
        '__ge__': [('__le__', lambda self, other: (not self >= other) or self == other),
                   ('__gt__', lambda self, other: self >= other and not self == other),
                   ('__lt__', lambda self, other: not self >= other)]
    }
    roots = set(dir(cls)) & set(convert)
    if not roots:
        raise ValueError('must define at least one ordering operation: < > <= >=')
    root = max(roots)       # prefer __lt__ to __le__ to __gt__ to __ge__
    for opname, opfunc in convert[root]:
        if opname not in roots:
            opfunc.__name__ = opname
            opfunc.__doc__ = getattr(int, opname).__doc__
            setattr(cls, opname, opfunc)
    return cls

def cmp_to_key(mycmp):
    """Convert a cmp= function into a key= function"""
    class K(object):
        __slots__ = ['obj']
        def __init__(self, obj):
            self.obj = obj
        def __lt__(self, other):
            return mycmp(self.obj, other.obj) < 0
        def __gt__(self, other):
            return mycmp(self.obj, other.obj) > 0
        def __eq__(self, other):
            return mycmp(self.obj, other.obj) == 0
        def __le__(self, other):
            return mycmp(self.obj, other.obj) <= 0
        def __ge__(self, other):
            return mycmp(self.obj, other.obj) >= 0
        def __ne__(self, other):
            return mycmp(self.obj, other.obj) != 0
        __hash__ = None
    return K

_CacheInfo = namedtuple("CacheInfo", "hits misses maxsize currsize")

def lru_cache(maxsize=100):
    """Least-recently-used cache decorator.

    If *maxsize* is set to None, the LRU features are disabled and the cache
    can grow without bound.

    Arguments to the cached function must be hashable.

    View the cache statistics named tuple (hits, misses, maxsize, currsize) with
    f.cache_info().  Clear the cache and statistics with f.cache_clear().
    Access the underlying function with f.__wrapped__.

    See:  http://en.wikipedia.org/wiki/Cache_algorithms#Least_Recently_Used

    """
    # Users should only access the lru_cache through its public API:
    #       cache_info, cache_clear, and f.__wrapped__
    # The internals of the lru_cache are encapsulated for thread safety and
    # to allow the implementation to change (including a possible C version).

    def decorating_function(user_function,
                tuple=tuple, sorted=sorted, len=len, KeyError=KeyError):

        hits, misses = [0], [0]
        kwd_mark = (object(),)          # separates positional and keyword args
        lock = Lock()                   # needed because OrderedDict isn't threadsafe

        if maxsize is None:
            cache = dict()              # simple cache without ordering or size limit

            @wraps(user_function)
            def wrapper(*args, **kwds):
                key = args
                if kwds:
                    key += kwd_mark + tuple(sorted(kwds.items()))
                try:
                    result = cache[key]
                    hits[0] += 1
                    return result
                except KeyError:
                    pass
                result = user_function(*args, **kwds)
                cache[key] = result
                misses[0] += 1
                return result
        else:
            cache = OrderedDict()           # ordered least recent to most recent
            cache_popitem = cache.popitem
            cache_renew = cache.move_to_end

            @wraps(user_function)
            def wrapper(*args, **kwds):
                key = args
                if kwds:
                    key += kwd_mark + tuple(sorted(kwds.items()))
                with lock:
                    try:
                        result = cache[key]
                        cache_renew(key)    # record recent use of this key
                        hits[0] += 1
                        return result
                    except KeyError:
                        pass
                result = user_function(*args, **kwds)
                with lock:
                    cache[key] = result     # record recent use of this key
                    misses[0] += 1
                    if len(cache) > maxsize:
                        cache_popitem(0)    # purge least recently used cache entry
                return result

        def cache_info():
            """Report cache statistics"""
            with lock:
                return _CacheInfo(hits[0], misses[0], maxsize, len(cache))

        def cache_clear():
            """Clear the cache and cache statistics"""
            with lock:
                cache.clear()
                hits[0] = misses[0] = 0

        wrapper.cache_info = cache_info
        wrapper.cache_clear = cache_clear
        return wrapper

    return decorating_function



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import Operand
from .base import HasInput
from .. import opcodes as OperandDef
from ..serialize import ValueType, KeyField, Int32Field, Int64Field, Float64Field, StringField, ListField


class DiscreteFourierTransform(HasInput):
    __slots__ = ()


class StandardFFT(DiscreteFourierTransform):
    _input = KeyField('input')
    _n = Int64Field('n')
    _axis = Int32Field('axis')
    _norm = StringField('norm')

    @property
    def n(self):
        return self._n

    @property
    def axis(self):
        return self._axis

    @property
    def norm(self):
        return getattr(self, '_norm', None)


class FFT(StandardFFT):
    _op_type_ = OperandDef.FFT


class IFFT(StandardFFT):
    _op_type_ = OperandDef.IFFT


class StandardFFTN(DiscreteFourierTransform):
    _input = KeyField('input')
    _shape = ListField('shape', ValueType.int64)
    _axes = ListField('axes', ValueType.int32)
    _norm = StringField('norm')

    @property
    def shape(self):
        return self._shape

    @property
    def axes(self):
        return self._axes

    @property
    def norm(self):
        return getattr(self, '_norm', None)


class FFT2(StandardFFTN):
    _op_type_ = OperandDef.FFT2


class IFFT2(StandardFFTN):
    _op_type_ = OperandDef.IFFT2


class FFTN(StandardFFTN):
    _op_type_ = OperandDef.FFTN


class IFFTN(StandardFFTN):
    _op_type_ = OperandDef.IFFTN


class RealFFT(DiscreteFourierTransform):
    _input = KeyField('input')
    _n = Int64Field('n')
    _axis = Int32Field('axis')
    _norm = StringField('norm')

    @property
    def n(self):
        return self._n

    @property
    def axis(self):
        return self._axis

    @property
    def norm(self):
        return getattr(self, '_norm', None)


class RFFT(RealFFT):
    _op_type_ = OperandDef.RFFT


class IRFFT(RealFFT):
    _op_type_ = OperandDef.IRFFT


class RealFFTN(DiscreteFourierTransform):
    _input = KeyField('input')
    _shape = ListField('shape', ValueType.int64)
    _axes = ListField('axes', ValueType.int32)
    _norm = StringField('norm')

    @property
    def shape(self):
        return self._shape

    @property
    def axes(self):
        return self._axes

    @property
    def norm(self):
        return getattr(self, '_norm', None)


class RFFT2(RealFFTN):
    _op_type_ = OperandDef.RFFT2


class IRFFT2(RealFFTN):
    _op_type_ = OperandDef.IRFFT2


class RFFTN(RealFFTN):
    _op_type_ = OperandDef.RFFTN


class IRFFTN(RealFFTN):
    _op_type_ = OperandDef.IRFFTN


class HermitianFFT(DiscreteFourierTransform):
    _input = KeyField('input')
    _n = Int64Field('n')
    _axis = Int32Field('axis')
    _norm = StringField('norm')

    @property
    def n(self):
        return self._n

    @property
    def axis(self):
        return self._axis

    @property
    def norm(self):
        return getattr(self, '_norm', None)


class HFFT(HermitianFFT):
    _op_type_ = OperandDef.HFFT


class IHFFT(HermitianFFT):
    _op_type_ = OperandDef.IHFFT


class FFTFreq(Operand):
    _op_type_ = OperandDef.FFTFREQ

    _n = Int32Field('n')
    _d = Float64Field('d')

    @property
    def n(self):
        return self._n

    @property
    def d(self):
        return self._d


class FFTFreqChunk(HasInput):
    _op_type_ = OperandDef.FFTFREQ_CHUNK

    _input = KeyField('input')
    _n = Int32Field('n')
    _d = Float64Field('d')

    @property
    def n(self):
        return self._n

    @property
    def d(self):
        return self._d


class RFFTFreq(Operand):
    _op_type_ = OperandDef.RFFTFREQ

    _n = Int32Field('n')
    _d = Float64Field('d')

    @property
    def n(self):
        return self._n

    @property
    def d(self):
        return self._d


class FFTShiftBase(HasInput):
    _input = KeyField('input')
    _axes = ListField('axes', ValueType.int32)

    @property
    def axes(self):
        return self._axes


class FFTShift(FFTShiftBase):
    _op_type_ = OperandDef.FFTSHIFT


class IFFTShift(FFTShiftBase):
    _op_type_ = OperandDef.IFFTSHIFT



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import operands
from .core import TensorRandomOperandMixin, handle_array


class TensorExponential(operands.Exponential, TensorRandomOperandMixin):
    __slots__ = '_scale', '_size'
    _input_fields_ = ['_scale']

    def __init__(self, state=None, size=None, dtype=None, gpu=None, **kw):
        dtype = np.dtype(dtype) if dtype is not None else dtype
        super(TensorExponential, self).__init__(_state=state, _size=size,
                                                _dtype=dtype, _gpu=gpu, **kw)

    def __call__(self, scale, chunks=None):
        return self.new_tensor([scale], self._size, raw_chunks=chunks)


def exponential(random_state, scale=1.0, size=None, chunks=None, gpu=None, **kw):
    """
    Draw samples from an exponential distribution.

    Its probability density function is

    .. math:: f(x; \frac{1}{\beta}) = \frac{1}{\beta} \exp(-\frac{x}{\beta}),

    for ``x > 0`` and 0 elsewhere. :math:`\beta` is the scale parameter,
    which is the inverse of the rate parameter :math:`\lambda = 1/\beta`.
    The rate parameter is an alternative, widely used parameterization
    of the exponential distribution [3]_.

    The exponential distribution is a continuous analogue of the
    geometric distribution.  It describes many common situations, such as
    the size of raindrops measured over many rainstorms [1]_, or the time
    between page requests to Wikipedia [2]_.

    Parameters
    ----------
    scale : float or array_like of floats
        The scale parameter, :math:`\beta = 1/\lambda`.
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  If size is ``None`` (default),
        a single value is returned if ``scale`` is a scalar.  Otherwise,
        ``np.array(scale).size`` samples are drawn.
    chunks : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    out : Tensor or scalar
        Drawn samples from the parameterized exponential distribution.

    References
    ----------
    .. [1] Peyton Z. Peebles Jr., "Probability, Random Variables and
           Random Signal Principles", 4th ed, 2001, p. 57.
    .. [2] Wikipedia, "Poisson process",
           http://en.wikipedia.org/wiki/Poisson_process
    .. [3] Wikipedia, "Exponential distribution",
           http://en.wikipedia.org/wiki/Exponential_distribution
    """
    if 'dtype' not in kw:
        kw['dtype'] = np.random.RandomState().exponential(
            handle_array(scale), size=(0,)).dtype
    size = random_state._handle_size(size)
    op = TensorExponential(state=random_state._state, size=size, gpu=gpu, **kw)
    return op(scale, chunks=chunks)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2018 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import contextlib

import numpy as np

from ....compat import lrange
from ..utils import unify_chunks, broadcast_shape, check_out_param
from ...core import Tensor, ExecutableTuple
from ..core import TensorOperandMixin
from ..datasource import tensor as astensor


class TensorElementWise(TensorOperandMixin):
    __slots__ = ()

    @classmethod
    def tile(cls, op):
        inputs = unify_chunks(*[(input, lrange(input.ndim)[::-1]) for input in op.inputs])

        chunk_shapes = [t.chunk_shape for t in inputs]
        out_chunk_shape = broadcast_shape(*chunk_shapes)

        out_chunks = [list() for _ in op.outputs]
        nsplits = [[None] * shape for shape in out_chunk_shape]
        get_index = lambda idx, t: tuple(0 if t.nsplits[i] == (1,) else ix for i, ix in enumerate(idx))
        for out_index in itertools.product(*(map(range, out_chunk_shape))):
            in_chunks = [t.cix[get_index(out_index[-t.ndim:], t)] if t.ndim != 0 else t.chunks[0]
                         for t in inputs]
            chunk_op = op.copy().reset_key()
            chunk_shape = broadcast_shape(*(c.shape for c in in_chunks))
            chunks = chunk_op.new_chunks(in_chunks, chunk_shape, index=out_index,
                                         dtype=[o.dtype for o in op.outputs],
                                         kws=[{'side': str(i)} for i in range(len(op.outputs))])
            for i, out_chunk in enumerate(chunks):
                out_chunks[i].append(out_chunk)
            for i, idx, s in zip(itertools.count(0), out_index, chunks[0].shape):
                nsplits[i][idx] = s

        new_op = op.copy()
        kws = [{'chunks': out_chunk, 'nsplits': nsplits} for out_chunk in out_chunks]
        return new_op.new_tensors(list(inputs), [o.shape for o in op.outputs],
                                  dtype=[o.dtype for o in op.outputs],
                                  kws=kws, output_limit=len(op.outputs))


class TensorBinOp(TensorElementWise):
    __slots__ = ()

    def check_inputs(self, inputs):
        if len(inputs) < 2 or len(inputs) > 4:
            raise ValueError(
                "Binary operand can only accept 2 to 4 inputs, got {0}".format(len(inputs)))

    @classmethod
    def _is_sparse(cls, x1, x2):
        return False

    @staticmethod
    def _process_inputs(x1, x2, out, where):
        x1, x2 = astensor(x1), astensor(x2)

        if out is not None and not isinstance(out, Tensor):
            raise TypeError('out should be Tensor object, got {0} instead'.format(type(out)))
        if where is True:
            where = None
        if where is not None:
            where = astensor(where)

        return x1, x2, out, where

    @classmethod
    def constant_cls(cls):
        raise NotImplementedError

    def to_constant(self, x1, x2):
        constant_op_cls = self.constant_cls()
        constant_op = constant_op_cls(getattr(self, '_casting'), getattr(self, '_err'),
                                      getattr(self, '_dtype'), getattr(self, '_sparse'))
        return constant_op(x1, x2)

    @contextlib.contextmanager
    def _handle_params(self, inputs):
        inps = inputs[:2]  # lhs, rhs
        inputs_iter = iter(inputs[2:])

        if getattr(self, '_lhs', None) is None:
            # create binop from beginning
            has_out = False
            out = next(inputs_iter)
            if out is not None:
                has_out = True
                inps.append(out)

            has_where = False
            where = next(inputs_iter)
            if where is not None:
                has_where = True
                inps.append(where)
        else:
            # create from exist binop
            has_out = getattr(self, '_out', None) is not None
            if has_out:
                out = next(inputs_iter)
                inps.append(out)

            has_where = getattr(self, '_where', None) is not None
            if has_where:
                where = next(inputs_iter)
                inps.append(where)

        yield inps

        inputs = getattr(self, '_inputs')
        inputs_iter = iter(inputs)
        setattr(self, '_lhs', next(inputs_iter))
        setattr(self, '_rhs', next(inputs_iter))
        if has_out:
            setattr(self, '_out', next(inputs_iter))
        if has_where:
            setattr(self, '_where', next(inputs_iter))

    def new_tensors(self, inputs, shape, **kw):
        with self._handle_params(inputs) as inputs:
            return super(TensorBinOp, self).new_tensors(inputs, shape, **kw)

    def new_chunks(self, inputs, shape, **kw):
        with self._handle_params(inputs) as inputs:
            return super(TensorBinOp, self).new_chunks(inputs, shape, **kw)

    def _call(self, x1, x2, out=None, where=None):
        # if x1 or x2 is scalar, and out is none, to constant
        if (np.isscalar(x1) or np.isscalar(x2)) and not out:
            return self.to_constant(x1, x2)

        x1, x2, out, where = self._process_inputs(x1, x2, out, where)
        # check broadcast
        shape = broadcast_shape(x1.shape, x2.shape)

        setattr(self, '_sparse', self._is_sparse(x1, x2))
        t = self.new_tensor([x1, x2, out, where], shape)

        if out is None:
            return t

        check_out_param(out, t, getattr(self, '_casting'))
        out_shape, out_dtype = out.shape, out.dtype

        # if `out` is specified, use out's dtype and shape
        if t.shape != out_shape:
            t = self.new_tensor([x1, x2, out, where], out_shape)
        setattr(self, '_dtype', out_dtype)

        out.data = t.data
        return out

    def __call__(self, x1, x2, out=None, where=None):
        return self._call(x1, x2, out=out, where=where)

    def rcall(self, x1, x2, out=None, where=None):
        return self._call(x2, x1, out=out, where=where)


class TensorConstant(TensorElementWise):
    __slots__ = ()

    @classmethod
    def _is_sparse(cls, x1, x2):
        return False

    @contextlib.contextmanager
    def _handle_params(self, inputs):
        inps = []
        if getattr(self, '_lhs', None) is None:
            # create a constant op from beginning
            assert len(inputs) == 2
            lhs_scalar = np.isscalar(inputs[0])
            rhs_scalar = np.isscalar(inputs[1])
            if not lhs_scalar:
                inps.append(inputs[0])
            else:
                setattr(self, '_lhs', inputs[0])
            if not rhs_scalar:
                inps.append(inputs[1])
            else:
                setattr(self, '_rhs', inputs[1])
            assert len(inps) <= 1  # all inputs are constant, or only 1 is constant
        else:
            # create from a exist constant op
            lhs_scalar = np.isscalar(getattr(self, '_lhs'))
            rhs_scalar = np.isscalar(getattr(self, '_rhs'))
            inps.extend(inp for inp in inputs if not np.isscalar(inp))

        yield inps

        inputs = getattr(self, '_inputs')
        inputs_iter = iter(inputs)
        if not lhs_scalar:
            setattr(self, '_lhs', next(inputs_iter))
        if not rhs_scalar:
            setattr(self, '_rhs', next(inputs_iter))

    def new_tensors(self, inputs, shape, **kw):
        with self._handle_params(inputs) as inputs:
            return super(TensorConstant, self).new_tensors(inputs, shape, **kw)

    def new_chunks(self, inputs, shape, **kw):
        with self._handle_params(inputs) as inputs:
            return super(TensorConstant, self).new_chunks(inputs, shape, **kw)

    def _call(self, x1, x2):
        x1_scalar = np.isscalar(x1)
        x2_scalar = np.isscalar(x2)
        if x1_scalar and x2_scalar:
            shape = ()
        elif x1_scalar:
            x2 = astensor(x2)
            shape = x2.shape
        else:
            x1 = astensor(x1)
            shape = x1.shape

        setattr(self, '_sparse', self._is_sparse(x1, x2))
        return self.new_tensor([x1, x2], shape)

    def __call__(self, x1, x2):
        return self._call(x1, x2)

    def rcall(self, x1, x2):
        return self._call(x2, x1)


class TensorUnaryOp(TensorElementWise):
    __slots__ = ()

    @staticmethod
    def _process_inputs(x, out, where):
        x = astensor(x)

        if out is not None and not isinstance(out, Tensor):
            raise TypeError('out should be Tensor object, got {0} instead'.format(type(out)))
        if where is True:
            where = None
        if where is not None:
            where = astensor(where)

        return x, out, where

    @classmethod
    def _is_sparse(cls, x):
        return False

    @contextlib.contextmanager
    def _handle_params(self, inputs):
        inps = inputs[:1]
        inputs_iter = iter(inputs[1:])

        if getattr(self, '_input', None) is None:
            # create unaryop from beginning
            has_out = False
            out = next(inputs_iter)
            if out is not None:
                has_out = True
                inps.append(out)

            has_where = False
            where = next(inputs_iter)
            if where is not None:
                has_where = True
                inps.append(where)
        else:
            # create from exist unaryop
            has_out = getattr(self, '_out', None) is not None
            if has_out:
                out = next(inputs_iter)
                inps.append(out)

            has_where = getattr(self, '_where', None) is not None
            if has_where:
                where = next(inputs_iter)
                inps.append(where)

        yield inps

        inputs = getattr(self, '_inputs')
        inputs_iter = iter(inputs)
        setattr(self, '_input', next(inputs_iter))
        if has_out:
            setattr(self, '_out', next(inputs_iter))
        if has_where:
            setattr(self, '_where', next(inputs_iter))

    def new_tensors(self, inputs, shape,**kw):
        with self._handle_params(inputs) as inputs:
            return super(TensorUnaryOp, self).new_tensors(inputs, shape, **kw)

    def new_chunks(self, inputs, shape, **kw):
        with self._handle_params(inputs) as inputs:
            return super(TensorUnaryOp, self).new_chunks(inputs, shape, **kw)

    def _call(self, x, out=None, where=None):
        x, out, where = self._process_inputs(x, out, where)
        shape = x.shape

        setattr(self, '_sparse', self._is_sparse(x))
        t = self.new_tensor([x, out, where], shape)

        if out is None:
            return t

        check_out_param(out, t, getattr(self, '_casting'))
        out_shape, out_dtype = out.shape, out.dtype

        # if `out` is specified, use out's dtype and shape
        if t.shape != out_shape:
            t = self.new_tensor([x, out, where], out_shape)
        setattr(self, '_dtype', out_dtype)

        out.data = t.data
        return out

    def __call__(self, x, out=None, where=None):
        return self._call(x, out=out, where=where)


class TensorCompare(TensorBinOp):
    @classmethod
    def _is_sparse(cls, x1, x2):
        if x1.issparse() and x2.issparse():
            return True
        return False


class TensorCompareConstant(TensorConstant):
    @classmethod
    def _is_sparse(cls, x1, x2):
        if hasattr(x1, 'issparse') and x1.issparse():
            return True
        if hasattr(x2, 'issparse') and x2.issparse():
            return True
        return False


class TensorOutBinOp(TensorElementWise):
    __slots__ = ()

    @staticmethod
    def _process_inputs(x, out1, out2, where):
        x = astensor(x)

        if out1 is not None and not isinstance(out1, Tensor):
            raise TypeError('out1 should be Tensor object, got {0} instead'.format(type(out1)))
        if out2 is not None and not isinstance(out2, Tensor):
            raise TypeError('out2 should be Tensor object, got {0} instead'.format(type(out2)))
        if where is True:
            where = None
        if where is not None:
            where = astensor(where)

        return x, out1, out2, where

    @classmethod
    def _is_sparse(cls, x):
        return False

    @contextlib.contextmanager
    def _handle_params(self, inputs):
        inps = inputs[:1]
        inputs_iter = iter(inputs[1:])

        if getattr(self, '_input', None) is None:
            # create op from biginning,
            has_out1 = False
            out1 = next(inputs_iter)
            if out1 is not None:
                has_out1 = True
                inps.append(out1)

            has_out2 = False
            out2 = next(inputs_iter)
            if out2 is not None:
                has_out2 = True
                inps.append(out2)

            has_where = False
            where = next(inputs_iter)
            if where is not None:
                has_where = True
                inps.append(where)
        else:
            # create from exist op
            has_out1 = getattr(self, '_out1', None) is not None
            if has_out1:
                out1 = next(inputs_iter)
                inps.append(out1)

            has_out2 = getattr(self, '_out2', None) is not None
            if has_out2:
                out2 = next(inputs_iter)
                inps.append(out2)

            has_where = getattr(self, '_where', None) is not None
            if has_where:
                where = next(inputs_iter)
                inps.append(where)

        yield inps

        inputs = getattr(self, '_inputs')
        inputs_iter = iter(inputs)
        setattr(self, '_input', next(inputs_iter))
        if has_out1:
            setattr(self, '_out1', next(inputs_iter))
        if has_out2:
            setattr(self, '_out2', next(inputs_iter))
        if has_where:
            setattr(self, '_where', next(inputs_iter))

    def new_tensors(self, inputs, shape, **kw):
        with self._handle_params(inputs) as inputs:
            return super(TensorOutBinOp, self).new_tensors(inputs, shape, **kw)

    def new_chunks(self, inputs, shape, **kw):
        with self._handle_params(inputs) as inputs:
            return super(TensorOutBinOp, self).new_chunks(inputs, shape, **kw)

    @property
    def _fun(self):
        raise NotImplementedError

    def _call(self, x, out1=None, out2=None, out=None, where=None):
        dtype = [r.dtype for r in self._fun(np.empty(1, dtype=x.dtype))]

        out = out or (None, None)
        out1 = out1 or out[0]
        out2 = out2 or out[1]
        x, out1, out2, where = self._process_inputs(x, out1, out2, where)
        shape = x.shape

        setattr(self, '_sparse', self._is_sparse(x))
        t1, t2 = self.new_tensors([x, out1, out2, where], shape, dtype=dtype)

        if out1 is None and out2 is None:
            return ExecutableTuple([t1, t2])

        if out1 is not None:
            check_out_param(out1, t1, getattr(self, '_casting'))
            out1_shape, out1_dtype = out1.shape, out1.dtype
        else:
            out1_shape, out1_dtype = t1.shape, t1.dtype
        if out2 is not None:
            check_out_param(out2, t2, getattr(self, '_casting'))
            out2_shape, out2_dtype = out2.shape, out2.dtype
        else:
            out2_shape, out2_dtype = t2.shape, t2.dtype
        # if `out` is specified, use out's dtype and shape
        if t1.shape != out1_shape or t2.shape != out2_shape:
            t1, t2 = self.new_tensor([x, out1, out2, where], [out1_shape, out2_shape],
                                     dtype=[out1_dtype, out2_dtype])

        if out1 is not None:
            out1.data = t1.data
        else:
            out1 = t1
        if out2 is not None:
            out2.data = t2.data
        else:
            out2 = t2
        return ExecutableTuple([out1, out2])

    def __call__(self, x, out1=None, out2=None, out=None, where=None):
        return self._call(x, out1=out1, out2=out2, out=out, where=where)
